from typing import Any, TypeVar

from hugr import ops
from hugr import tys as ht
from hugr.build.dfg import DfBase

from guppylang_internals.ast_util import AstNode
from guppylang_internals.checker.errors.comptime_errors import (
    IllegalComptimeExpressionError,
)
from guppylang_internals.checker.expr_checker import python_value_to_guppy_type
from guppylang_internals.compiler.core import CompilerContext
from guppylang_internals.compiler.expr_compiler import python_value_to_hugr
from guppylang_internals.error import GuppyComptimeError, GuppyError
from guppylang_internals.std._internal.compiler.array import array_new, unpack_array
from guppylang_internals.std._internal.compiler.prelude import build_unwrap
from guppylang_internals.tracing.frozenlist import frozenlist
from guppylang_internals.tracing.object import (
    GuppyObject,
    GuppyStructObject,
    TracingDefMixin,
)
from guppylang_internals.tracing.state import get_tracing_state
from guppylang_internals.tys.builtin import (
    array_type,
    get_array_length,
    get_element_type,
    is_array_type,
)
from guppylang_internals.tys.const import ConstValue
from guppylang_internals.tys.ty import NoneType, StructType, TupleType

P = TypeVar("P", bound=ops.DfParentOp)


def unpack_guppy_object(
    obj: GuppyObject, builder: DfBase[P], frozen: bool = False
) -> Any:
    """Tries to turn as much of the structure of a GuppyObject into Python objects.

    For example, Guppy tuples are turned into Python tuples and Guppy arrays are turned
    into Python lists. This is achieved by inserting unpacking operations into the Hugr
    to get individual wires to be used in those Python objects.

    Setting `frozen=True` ensures that the resulting Python objects are not mutable in-
    place. This should be set for objects that originate from function inputs that are
    not borrowed.
    """
    match obj._ty:
        case NoneType():
            return None
        case TupleType(element_types=tys):
            unpack = builder.add_op(ops.UnpackTuple(), obj._use_wire(None))
            return tuple(
                unpack_guppy_object(GuppyObject(ty, wire), builder, frozen)
                for ty, wire in zip(tys, unpack.outputs(), strict=True)
            )
        case StructType() as ty:
            unpack = builder.add_op(ops.UnpackTuple(), obj._use_wire(None))
            field_values = [
                unpack_guppy_object(GuppyObject(field.ty, wire), builder, frozen)
                for field, wire in zip(ty.fields, unpack.outputs(), strict=True)
            ]
            return GuppyStructObject(ty, field_values, frozen)
        case ty if is_array_type(ty):
            length = get_array_length(ty)
            if isinstance(length, ConstValue):
                if length.value == 0:
                    # Zero-length lists cannot be turned back ito Guppy objects since
                    # there is no way to infer the type. Therefore, we should leave
                    # them as Guppy objects here
                    return obj
                elem_ty = get_element_type(ty)
                opt_elems = unpack_array(builder, obj._use_wire(None))
                err = "Non-copyable array element has already been used"
                elems = [build_unwrap(builder, opt_elem, err) for opt_elem in opt_elems]
                obj_list = [
                    unpack_guppy_object(GuppyObject(elem_ty, wire), builder, frozen)
                    for wire in elems
                ]
                return frozenlist(obj_list) if frozen else obj_list
            else:
                # Cannot handle generic sizes
                return obj
        case _:
            return obj


def guppy_object_from_py(
    v: Any, builder: DfBase[P], node: AstNode, ctx: CompilerContext
) -> GuppyObject:
    """Constructs a Guppy object from a Python value.

    Essentially undoes the `unpack_guppy_object` operation.
    """
    match v:
        case GuppyObject() as obj:
            return obj
        case TracingDefMixin() as defn:
            return defn.to_guppy_object()
        case None:
            return GuppyObject(NoneType(), builder.add_op(ops.MakeTuple()))
        case tuple(vs):
            objs = [guppy_object_from_py(v, builder, node, ctx) for v in vs]
            return GuppyObject(
                TupleType([obj._ty for obj in objs]),
                builder.add_op(ops.MakeTuple(), *(obj._use_wire(None) for obj in objs)),
            )
        case GuppyStructObject(_ty=struct_ty, _field_values=values):
            wires = []
            for f in struct_ty.fields:
                obj = guppy_object_from_py(values[f.name], builder, node, ctx)
                # Check that the field still has the correct type. Since we allow users
                # to mutate structs unchecked, this needs to be checked here
                if obj._ty != f.ty:
                    raise GuppyComptimeError(
                        f"Field `{f.name}` of object with type `{struct_ty}` has an "
                        f"unexpected type. Expected `{f.ty}`, got `{obj._ty}`."
                    )
                wires.append(obj._use_wire(None))
            return GuppyObject(struct_ty, builder.add_op(ops.MakeTuple(), *wires))
        case list(vs) if len(vs) > 0:
            objs = [guppy_object_from_py(v, builder, node, ctx) for v in vs]
            elem_ty = objs[0]._ty
            for i, obj in enumerate(objs[1:]):
                if obj._ty != elem_ty:
                    raise GuppyComptimeError(
                        f"Element at index {i + 1} does not match the type of "
                        f"previous elements. Expected `{elem_ty}`, got `{obj._ty}`."
                    )
            hugr_elem_ty = ht.Option(elem_ty.to_hugr(ctx))
            wires = [
                builder.add_op(ops.Tag(1, hugr_elem_ty), obj._use_wire(None))
                for obj in objs
            ]
            return GuppyObject(
                array_type(elem_ty, len(vs)),
                builder.add_op(array_new(hugr_elem_ty, len(vs)), *wires),
            )
        case []:
            # Empty lists are tricky since we can't infer the element type here
            # TODO: Propagate type information?
            raise GuppyComptimeError("Cannot infer the type of empty list")
        case v:
            ty = python_value_to_guppy_type(v, node, get_tracing_state().globals)
            if ty is None:
                raise GuppyError(IllegalComptimeExpressionError(node, type(v)))
            hugr_val = python_value_to_hugr(v, ty, ctx)
            assert hugr_val is not None
            return GuppyObject(ty, builder.load(hugr_val))


def update_packed_value(v: Any, obj: "GuppyObject", builder: DfBase[P]) -> None:
    """Given a Python value `v` and a `GuppyObject` `obj` that was constructed from `v`
    using `guppy_object_from_py`, updates the wires of any `GuppyObjects` contained in
    `v` to the new wires specified by `obj`.

    Also resets the used flag on any of those updated wires. This corresponds to making
    the object available again since it now corresponds to a fresh wire.
    """
    match v:
        case GuppyObject() as v_obj:
            assert v_obj._ty == obj._ty
            v_obj._wire = obj._use_wire(None)
            if not v_obj._ty.droppable and v_obj._used:
                state = get_tracing_state()
                state.unused_undroppable_objs[v_obj._id] = v_obj
            v_obj._used = None
        case None:
            assert isinstance(obj._ty, NoneType)
        case tuple(vs):
            assert isinstance(obj._ty, TupleType)
            wires = builder.add_op(ops.UnpackTuple(), obj._use_wire(None)).outputs()
            for v, ty, wire in zip(vs, obj._ty.element_types, wires, strict=True):
                update_packed_value(v, GuppyObject(ty, wire), builder)
        case GuppyStructObject(_ty=ty, _field_values=values):
            assert obj._ty == ty
            wires = builder.add_op(ops.UnpackTuple(), obj._use_wire(None)).outputs()
            for (
                field,
                wire,
            ) in zip(ty.fields, wires, strict=True):
                v = values[field.name]
                update_packed_value(v, GuppyObject(field.ty, wire), builder)
        case list(vs) if len(vs) > 0:
            assert is_array_type(obj._ty)
            elem_ty = get_element_type(obj._ty)
            opt_wires = unpack_array(builder, obj._use_wire(None))
            err = "Non-droppable array element has already been used"
            for v, opt_wire in zip(vs, opt_wires, strict=True):
                (wire,) = build_unwrap(builder, opt_wire, err).outputs()
                update_packed_value(v, GuppyObject(elem_ty, wire), builder)
        case _:
            pass



import ast
import itertools
from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar, TypeAlias

from hugr.build.dfg import DefinitionBuilder, OpVar

from guppylang_internals.diagnostic import Fatal
from guppylang_internals.span import SourceMap

if TYPE_CHECKING:
    from guppylang_internals.checker.core import Globals
    from guppylang_internals.compiler.core import (
        CompilerContext,
        PartiallyMonomorphizedArgs,
    )
    from guppylang_internals.tys.param import Parameter


RawDef: TypeAlias = "ParsableDef | ParsedDef"
ParsedDef: TypeAlias = "CheckableDef | CheckedDef"
CheckedDef: TypeAlias = "CompilableDef | MonomorphizableDef | CompiledDef"


@dataclass(frozen=True)
class DefId:
    """Unique identifier for definitions across modules.

    This id is persistent across all compilation stages. It can be used to identify a
    definition at any step in the compilation pipeline.

    Args:
        id: An integer uniquely identifying the definition.
        module: The module where the definition was defined.
    """

    id: int

    _ids: ClassVar[Iterator[int]] = itertools.count()

    @classmethod
    def fresh(cls) -> "DefId":
        return DefId(next(cls._ids))


@dataclass(frozen=True)
class Definition(ABC):
    """Abstract base class for user-defined objects on module-level.

    Each definition is identified by a globally unique id. Furthermore, we store the
    user-picked name for the defined object and an optional AST node for the definition
    location.

    Args:
        id: The unique definition identifier.
        name: The name of the definition.
        defined_at: The AST node where the definition was defined.
    """

    id: DefId
    name: str
    defined_at: ast.AST | None

    @property
    @abstractmethod
    def description(self) -> str:
        """Description of this definition to be used in messages to the user.

        The returned text should fit into messages of the following form: "expected
        a function, but got {description of this definition} instead".
        """


class ParsableDef(Definition):
    """Abstract base class for raw definitions that still require parsing.

    For example, raw function definitions first need to parse their signature and check
    that all types are valid. The result of parsing should be a definition that is ready
    to be checked.

    Args:
        id: The unique definition identifier.
        name: The name of the definition.
        defined_at: The AST node where the definition was defined.
    """

    @abstractmethod
    def parse(self, globals: "Globals", sources: SourceMap) -> ParsedDef:
        """Performs parsing and validation, returning a definition that can be checked.

        The provided globals contain all other raw definitions that have been defined.
        """


class CheckableDef(Definition):
    """Abstract base class for definitions that still need to be checked.

    The result of checking should be a definition that is ready to be compiled to Hugr.

    Args:
        id: The unique definition identifier.
        name: The name of the definition.
        defined_at: The AST node where the definition was defined.
    """

    @abstractmethod
    def check(self, globals: "Globals") -> CheckedDef:
        """Type checks the definition.

        The provided globals contain all other parsed definitions that have been
        defined.

        Returns a checked version of the definition that can be compiled to Hugr.
        """


class CompilableDef(Definition):
    """Abstract base class for definitions that still need to be compiled to Hugr.

    The result of compilation should be a `CompiledDef` with a pointer to the Hugr node
    that was created for this definition.

    Args:
        id: The unique definition identifier.
        name: The name of the definition.
        defined_at: The AST node where the definition was defined.
    """

    @abstractmethod
    def compile_outer(
        self, module: DefinitionBuilder[OpVar], ctx: "CompilerContext"
    ) -> "CompiledDef":
        """Adds a Hugr node for the definition to the provided Hugr module.

        Note that is not required to fill in the contents of the node. At this point,
        we don't have access to the globals since they have not all been compiled yet.

        See `CompiledDef.compile_inner()` for the hook to compile the inside of the
        node. This two-step process enables things like mutual recursion.
        """


class MonomorphizableDef(Definition):
    """Abstract base class for definitions that require monomorphization when compiling
    to Hugr.

    Args:
        id: The unique definition identifier.
        name: The name of the definition.
        defined_at: The AST node where the definition was defined.
    """

    @property
    @abstractmethod
    def params(self) -> "Sequence[Parameter]":
        """Generic parameters of this definition."""

    @abstractmethod
    def monomorphize(
        self,
        module: DefinitionBuilder[OpVar],
        mono_args: "PartiallyMonomorphizedArgs",
        ctx: "CompilerContext",
    ) -> "MonomorphizedDef":
        """Adds a Hugr node for the (partially) monomorphized definition to the provided
        Hugr module.

        See `MonomorphizedDef.compile_inner()` for the hook to compile the inside of the
        node. This two-step process enables things like mutual recursion.
        """


class CompiledDef(Definition):
    """Abstract base class for definitions that have been added to a Hugr.

    Args:
        id: The unique definition identifier.
        name: The name of the definition.
        defined_at: The AST node where the definition was defined.
    """

    def compile_inner(self, ctx: "CompilerContext") -> None:
        """Optional hook that is called to fill in the content of the Hugr node.

        Opposed to `CompilableDef.compile()`, we have access to all other compiled
        definitions here, which allows things like mutual recursion.
        """


@dataclass(frozen=True)
class MonomorphizedDef(CompiledDef):
    """Abstract base class for definitions that have been added to a Hugr and have been
    partially monomorphized.

    Args:
        id: The unique definition identifier.
        name: The name of the definition. This will be the same for all monomorphized
            variants of the definition.
        defined_at: The AST node where the original polymorphic definition was defined.
        mono_args: Partial monomorphization of the generic parameters.
    """

    mono_args: "PartiallyMonomorphizedArgs"


@dataclass(frozen=True)
class UnknownSourceError(Fatal):
    title: ClassVar[str] = "Cannot find source"
    message: ClassVar[str] = (
        "Unable to look up the source code for Python object `{obj}`"
    )
    obj: object



from guppylang.decorator import guppy
from guppylang.std.builtins import array


@guppy.declare
def foo(xs: array[int, 0]) -> None: ...


@guppy.comptime
def test(xs: array[int, 0]) -> None:
    foo(xs)  # This works
    foo([])  # But we currently cannot infer the type here


test.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import array, owned


@guppy.comptime
def foo(xs: array[int, 1] @ owned) -> None:
    xs[0] = 1


@guppy.comptime
def main() -> None:
    xs = [0]
    foo(xs)
    # Python users would expect xs to be mutated by the call above. However, we cannot
    # provide this semantics, so the mutation in `foo` is rejected by the compiler.


main.compile()



from typing import Any, TypeVar


class _DummyGuppy:
    """A dummy class with the same interface as `@guppy` that does nothing when used to
    decorate functions.

    We use this during sphinx builds as a mock for the decorator, to ensure that Guppy
    functions are recognised as regular functions and included in docs.
    """

    def __call__(self, f: Any) -> Any:
        return f

    def comptime(self, f: Any) -> Any:
        return f

    def extend_type(self, *args: Any, **kwargs: Any) -> Any:
        return lambda cls: cls

    def type(self, *args: Any, **kwargs: Any) -> Any:
        return lambda cls: cls

    def struct(self, cls: Any) -> Any:
        return cls

    def type_var(self, name: str, *args: Any, **kwargs: Any) -> Any:
        return TypeVar(name)

    def nat_var(self, name: str, *args: Any, **kwargs: Any) -> Any:
        return TypeVar(name)

    def custom(self, *args: Any, **kwargs: Any) -> Any:
        return lambda f: f

    def hugr_op(self, *args: Any, **kwargs: Any) -> Any:
        return lambda f: f

    def declare(self, f: Any) -> Any:
        return f

    def overload(self, *args: Any, **kwargs: Any) -> Any:
        return lambda f: f

    def constant(self, *args: Any, **kwargs: Any) -> Any:
        return None

    def extern(self, *args: Any, **kwargs: Any) -> Any:
        return None

    def check(self, *args: Any, **kwargs: Any) -> None:
        pass

    def compile(self, *args: Any, **kwargs: Any) -> Any:
        return None

    def compile_function(self, *args: Any, **kwargs: Any) -> Any:
        return None

    def pytket(self, *args: Any, **kwargs: Any) -> Any:
        return lambda f: f

    def load_pytket(self, *args: Any, **kwargs: Any) -> Any:
        return None


def _dummy_custom_decorator(*args: Any, **kwargs: Any) -> Any:
    """Dummy version of custom decorators that are used during Sphinx builds."""
    return lambda f: f


def sphinx_running() -> bool:
    """Checks if guppylang was imported during a sphinx build."""
    # This is the most general solution available at the moment.
    # See: https://github.com/sphinx-doc/sphinx/issues/9805
    try:
        import sphinx  # type: ignore[import-untyped, import-not-found, unused-ignore]

        return hasattr(sphinx, "application")
    except ImportError:
        return False



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit

T = guppy.type_var("T")


@guppy.declare
def foo(x: T) -> None:
    ...


@guppy
def main(q: qubit) -> None:
    foo(q)


main.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned, array
from guppylang.std.quantum import qubit, measure


@guppy.comptime
def test(qs: array[qubit, 10] @ owned) -> None:
    for i in range(9):
        measure(qs[i])


test.compile()



from guppylang import guppy
from guppylang_internals.decorator import custom_type
from guppylang_internals.tys.ty import NoneType

@custom_type(NoneType())
class MyType:
    """A non-iterable type."""


@guppy
def test(x: MyType) -> None:
    for _ in x:
        pass


test.compile()



"""Utilities for fixed-size arrays, denoted `array[T, n]`, for element type `T` and
compile-time constant size `n`.

See `frozenarray[T, n]` for an immutable version of the `array[T, n]` type.
"""

# mypy: disable-error-code="empty-body, misc, override, valid-type, no-untyped-def"

from __future__ import annotations

import builtins
from types import GeneratorType
from typing import TYPE_CHECKING, Generic, TypeVar, no_type_check

from guppylang_internals.decorator import custom_function, extend_type
from guppylang_internals.definition.custom import CopyInoutCompiler
from guppylang_internals.std._internal.checker import ArrayCopyChecker, NewArrayChecker
from guppylang_internals.std._internal.compiler.array import (
    ArrayGetitemCompiler,
    ArrayIterAsertAllUsedCompiler,
    ArraySetitemCompiler,
    NewArrayCompiler,
)
from guppylang_internals.std._internal.compiler.frozenarray import (
    FrozenarrayGetitemCompiler,
)
from guppylang_internals.tys.builtin import array_type_def, frozenarray_type_def

from guppylang import guppy
from guppylang.std.iter import SizedIter
from guppylang.std.option import Option, nothing, some

if TYPE_CHECKING:
    from guppylang.std.lang import owned


T = guppy.type_var("T")
L = guppy.type_var("L", copyable=False, droppable=False)
n = guppy.nat_var("n")

_T = TypeVar("_T")
_n = TypeVar("_n")


@extend_type(array_type_def)
class array(builtins.list[_T], Generic[_T, _n]):
    """Sequence of homogeneous values with statically known fixed length."""

    @custom_function(ArrayGetitemCompiler())
    def __getitem__(self: array[L, n], idx: int) -> L: ...

    @custom_function(ArraySetitemCompiler())
    def __setitem__(self: array[L, n], idx: int, value: L @ owned) -> None: ...

    @guppy
    @no_type_check
    def __len__(self: array[L, n]) -> int:
        return n

    @custom_function(NewArrayCompiler(), NewArrayChecker(), higher_order_value=False)
    def __new__(): ...

    # `__new__` will be overwritten below to provide actual runtime behaviour for
    # comptime. We still need to hold on to a reference to the Guppy function so
    # `@extend_type` can find it
    __new_guppy__ = __new__

    @guppy
    @no_type_check
    def __iter__(self: array[L, n] @ owned) -> SizedIter[ArrayIter[L, n], n]:
        return SizedIter(ArrayIter(self, 0))

    @custom_function(CopyInoutCompiler(), ArrayCopyChecker())
    def copy(self: array[T, n]) -> array[T, n]: ...

    def __new__(cls, *args: _T) -> builtins.list[_T]:  # type: ignore[no-redef]
        # Runtime array constructor that is used for comptime. We return an actual list
        # in line with the comptime unpacking logic that turns arrays into lists.
        if len(args) == 1 and isinstance(args[0], GeneratorType):
            return list(args[0])
        return [*args]


@guppy.struct
class ArrayIter(Generic[L, n]):
    """Iterator over arrays."""

    xs: array[L, n]
    i: int

    @guppy
    @no_type_check
    def __next__(
        self: ArrayIter[L, n] @ owned,
    ) -> Option[tuple[L, ArrayIter[L, n]]]:
        if self.i < int(n):
            elem = _array_unsafe_getitem(self.xs, self.i)
            return some((elem, ArrayIter(self.xs, self.i + 1)))
        self._assert_all_used()
        return nothing()

    @custom_function(ArrayIterAsertAllUsedCompiler())
    def _assert_all_used(self: ArrayIter[L, n] @ owned) -> None: ...


@custom_function(ArrayGetitemCompiler())
def _array_unsafe_getitem(xs: array[L, n], idx: int) -> L: ...


@extend_type(frozenarray_type_def)
class frozenarray(Generic[T, n]):
    """An immutable array of fixed static size."""

    @custom_function(FrozenarrayGetitemCompiler())
    def __getitem__(self: frozenarray[T, n], item: int) -> T: ...  # type: ignore[type-arg]

    @guppy
    @no_type_check
    def __len__(self: frozenarray[T, n]) -> int:
        return n

    @guppy
    @no_type_check
    def __iter__(self: frozenarray[T, n]) -> SizedIter[FrozenarrayIter[T, n], n]:
        return SizedIter(FrozenarrayIter(self, 0))

    @guppy
    @no_type_check
    def mutable_copy(self: frozenarray[T, n]) -> array[T, n]:
        """Creates a mutable copy of this array."""
        return array(x for x in self)


@guppy.struct
class FrozenarrayIter(Generic[T, n]):
    """Iterator for frozenarrays."""

    xs: frozenarray[T, n]  # type: ignore[type-arg]
    i: int

    @guppy
    @no_type_check
    def __next__(
        self: FrozenarrayIter[T, n],
    ) -> Option[tuple[T, FrozenarrayIter[T, n]]]:
        if self.i < int(n):
            return some((self.xs[self.i], FrozenarrayIter(self.xs, self.i + 1)))
        return nothing()



from guppylang import guppy
from guppylang.std.builtins import result, nat, array, comptime
from tests.util import compile_guppy


def test_basic(validate):
    @compile_guppy
    def main(x: int) -> None:
        result("foo", x)

    validate(main)


def test_multi(validate):
    @compile_guppy
    def main(w: nat, x: int, y: float, z: bool) -> None:
        result("a", w)
        result("b", x)
        result("c", y)
        result("d", z)

    validate(main)


def test_array(validate):
    @compile_guppy
    def main(
        w: array[nat, 42], x: array[int, 5], y: array[float, 1], z: array[bool, 0]
    ) -> None:
        result("a", w)
        result("b", x)
        result("c", y)
        result("d", z)

    validate(main)


def test_array_generic(validate):
    n = guppy.nat_var("n")

    @guppy
    def main(
        w: array[nat, n], x: array[int, n], y: array[float, n], z: array[bool, n]
    ) -> None:
        result("a", w)
        result("b", x)
        result("c", y)
        result("d", z)

    validate(main.compile())


def test_array_drop_after_result(validate):
    @compile_guppy
    def main() -> None:
        result("a", array(1, 2, 3))

    validate(main)


def test_same_tag(validate):
    @compile_guppy
    def main(x: int, y: float, z: bool) -> None:
        result("foo", x)
        result("foo", y)
        result("foo", z)

    validate(main)


def test_comptime_tag_inside(validate):
    @compile_guppy
    def main(x: int) -> None:
        result(comptime("a" + "b"), x)

    validate(main)


def test_comptime_tag_outside1(validate):
    EXAMPLE_RESULTS = [
        ("boolean", False),
        ("int", 123),
    ]

    @guppy.comptime
    def main() -> None:
        for key, value in EXAMPLE_RESULTS:
            result(key, value)

    validate(main.compile())


def test_comptime_tag_outside2(validate):
    EXAMPLE_RESULT = ("boolean", False)

    @guppy.comptime
    def main() -> None:
        result(EXAMPLE_RESULT[0], EXAMPLE_RESULT[1])

    validate(main.compile())



"""Tests Python 3.12 style generic syntax."""

from guppylang import array
from guppylang.decorator import guppy
from guppylang.std.lang import Copy, Drop, owned
from guppylang.std.num import nat
from guppylang.std.option import Option, nothing


def test_function(validate):
    @guppy
    def main[S, T](x: S @ owned, y: T @ owned) -> tuple[T, S]:
        return y, x

    validate(main.compile())


def test_struct(validate):
    @guppy.struct
    class MyStruct[S, T]:
        x: S
        y: T

    @guppy
    def main(s: MyStruct[int, float]) -> float:
        return s.x + s.y

    validate(main.compile())


def test_inner_frame(validate):
    """See https://github.com/CQCL/guppylang/issues/1116"""

    def make():
        @guppy.struct
        class MyStruct[T]:
            @guppy
            def foo(self: "MyStruct[int]") -> None:
                pass

        @guppy
        def main() -> None:
            MyStruct[int]().foo()

        return main

    validate(guppy.compile(make()))


def test_copy_bound(validate):
    @guppy.struct
    class MyStruct[T: Copy]:
        x: T

    @guppy
    def main[T: Copy](s: MyStruct[T]) -> tuple[T, T]:
        return s.x, s.x

    validate(main.compile())


def test_drop_bound(validate):
    @guppy.struct
    class MyStruct[T: Drop]:
        x: T

    @guppy
    def main[T: Drop](s: MyStruct[T] @ owned) -> None:
        pass

    validate(main.compile())


def test_copy_and_drop_bound(validate):
    @guppy.struct
    class MyStruct[T: (Copy, Drop)]:
        x: T

    @guppy
    def main[T: (Copy, Drop)](s1: MyStruct[T], s2: MyStruct[T]) -> tuple[T, T]:
        return s1.x, s1.x

    validate(main.compile())


def test_const_param(validate):
    @guppy.struct
    class MyStruct[T, n: nat]:
        xs: array[T, n]

    @guppy
    def main[T, n: nat](xs: array[T, n], s: MyStruct[T, n]) -> nat:
        return n

    validate(main.compile())


def test_mixed_legacy_params(validate):
    T = guppy.type_var("T", copyable=False, droppable=False)

    @guppy
    def main[S](x: S @ owned, y: T @ owned) -> tuple[T, S]:
        return y, x

    validate(main.compile())


def test_reference_inside(validate):
    @guppy
    def main[T: Drop]() -> None:
        x: Option[T] = nothing()
        nothing[T]()

    validate(main.compile())



from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar

from hugr import ops
from hugr.build.dfg import DfBase

from guppylang_internals.ast_util import AstNode, with_loc, with_type
from guppylang_internals.cfg.builder import tmp_vars
from guppylang_internals.checker.core import ComptimeVariable, Context, Locals, Variable
from guppylang_internals.checker.errors.type_errors import TypeMismatchError
from guppylang_internals.compiler.core import CompilerContext, DFContainer
from guppylang_internals.compiler.expr_compiler import ExprCompiler
from guppylang_internals.definition.value import CallableDef
from guppylang_internals.diagnostic import Error
from guppylang_internals.error import GuppyComptimeError, GuppyError, exception_hook
from guppylang_internals.nodes import PlaceNode
from guppylang_internals.tracing.builtins_mock import mock_builtins
from guppylang_internals.tracing.object import GuppyObject
from guppylang_internals.tracing.state import (
    TracingState,
    get_tracing_state,
    set_tracing_state,
)
from guppylang_internals.tracing.unpacking import (
    P,
    guppy_object_from_py,
    unpack_guppy_object,
    update_packed_value,
)
from guppylang_internals.tracing.util import capture_guppy_errors, tracing_except_hook
from guppylang_internals.tys.ty import FunctionType, InputFlags, type_to_row, unify

if TYPE_CHECKING:
    import ast

    from hugr import Wire


@dataclass(frozen=True)
class TracingReturnError(Error):
    title: ClassVar[str] = "Error in comptime function return"
    message: ClassVar[str] = "{msg}"
    msg: str


def trace_function(
    python_func: Callable[..., Any],
    ty: FunctionType,
    builder: DfBase[P],
    ctx: CompilerContext,
    node: AstNode,
) -> None:
    """Kicks off tracing of a function.

    Invokes the passed Python callable and constructs the corresponding Hugr using the
    passed builder.
    """
    state = TracingState(ctx, DFContainer(builder, ctx, {}), node)
    with set_tracing_state(state):
        inputs = [
            unpack_guppy_object(
                GuppyObject(inp.ty, wire),
                builder,
                # Function inputs are only allowed to be mutable if they are borrowed.
                # For owned arguments, mutation wouldn't be observable by the caller,
                # thus breaking the semantics expected from Python.
                frozen=InputFlags.Inout not in inp.flags,
            )
            for wire, inp in zip(builder.inputs(), ty.inputs, strict=True)
        ]

        with exception_hook(tracing_except_hook):
            mock_builtins(python_func)
            py_out = python_func(*inputs)

        try:
            out_obj = guppy_object_from_py(py_out, builder, node, ctx)
        except GuppyComptimeError as err:
            # Error in the return statement. For example, this happens if users
            # try to return a struct with invalid field values or there is a linearity
            # violation.
            raise GuppyError(TracingReturnError(node, str(err))) from None

        # Check that the output type is correct
        if unify(out_obj._ty, ty.output, {}) is None:
            raise GuppyError(
                TypeMismatchError(node, ty.output, out_obj._ty, "return value")
            )

        # Unpack regular returns
        out_tys = type_to_row(out_obj._ty)
        if len(out_tys) > 1:
            regular_returns: list[Wire] = list(
                builder.add_op(ops.UnpackTuple(), out_obj._use_wire(None)).outputs()
            )
        elif len(out_tys) > 0:
            regular_returns = [out_obj._use_wire(None)]
        else:
            regular_returns = []

        # Compute the inout extra outputs
        inout_returns = []
        assert ty.input_names is not None
        for inout_obj, inp, name in zip(inputs, ty.inputs, ty.input_names, strict=True):
            if InputFlags.Inout in inp.flags:
                err_prefix = (
                    f"Argument `{name}` is borrowed, so it is implicitly returned to "
                    f"the caller. "
                )
                try:
                    obj = guppy_object_from_py(inout_obj, builder, node, ctx)
                    inout_returns.append(obj._use_wire(None))
                except GuppyComptimeError as err:
                    msg = str(err)
                    if not msg.endswith("."):
                        msg += "."
                    e = TracingReturnError(node, err_prefix + msg)
                    raise GuppyError(e) from None
                # Also check that the type hasn't changed (for example, the user could
                # have changed the length of an array, thus changing its type)
                if obj._ty != inp.ty:
                    msg = (
                        f"{err_prefix}Expected it to have type `{inp.ty}`, but got "
                        f"`{obj._ty}`."
                    )
                    e = TracingReturnError(node, msg)
                    raise GuppyError(e) from None

    # Check that all allocated linear objects have been used
    if state.unused_undroppable_objs:
        _, unused = state.unused_undroppable_objs.popitem()
        msg = f"Value with non-droppable type `{unused._ty}` is leaked by this function"
        raise GuppyError(TracingReturnError(node, msg)) from None

    builder.set_outputs(*regular_returns, *inout_returns)


@capture_guppy_errors
def trace_call(func: CallableDef, *args: Any) -> Any:
    """Handles calls to Guppy functions during tracing.

    Checks that the passed arguments match the signature of the function and also
    handles inout arguments.
    """
    state = get_tracing_state()

    # Try to turn args into `GuppyObjects`
    args_objs = [
        guppy_object_from_py(arg, state.dfg.builder, state.node, state.ctx)
        for arg in args
    ]

    # Create dummy variables and bind the objects to them
    arg_vars: list[Variable] = [
        ComptimeVariable(next(tmp_vars), obj._ty, None, static_value=arg)
        for (obj, arg) in zip(args_objs, args, strict=True)
    ]
    locals = Locals({var.name: var for var in arg_vars})
    for obj, var in zip(args_objs, arg_vars, strict=True):
        state.dfg[var] = obj._use_wire(func)

    # Check call
    arg_exprs: list[ast.expr] = [
        with_loc(state.node, with_type(var.ty, PlaceNode(var))) for var in arg_vars
    ]
    call_node, ret_ty = func.synthesize_call(
        arg_exprs, state.node, Context(state.globals, locals, {})
    )

    # Compile call
    ret_wire = ExprCompiler(state.ctx).compile(call_node, state.dfg)

    # Update inouts
    # If the input types of the function aren't known, we can't check this.
    # This is the case for functions with a custom checker and no type annotations.
    if len(func.ty.inputs) != 0:
        for inp, arg, var in zip(func.ty.inputs, args, arg_vars, strict=True):
            if InputFlags.Inout in inp.flags:
                inout_wire = state.dfg[var]
                update_packed_value(
                    arg, GuppyObject(inp.ty, inout_wire), state.dfg.builder
                )

    ret_obj = GuppyObject(ret_ty, ret_wire)
    return unpack_guppy_object(ret_obj, state.dfg.builder)



from typing import Generic

from guppylang.decorator import guppy

T = guppy.type_var("T")
n = guppy.nat_var("n")

# We don't yet support const vars referencing previously bound parameters
X = guppy.const_var("X", "array[T, n]")


@guppy.struct
class Struct(Generic[T, n, X]):
    pass


@guppy
def main(_: Struct[T, n, X]) -> None:
    pass


main.compile()



from typing import Any

from typing_extensions import Self

from guppylang_internals.error import GuppyComptimeError

ERROR_MSG = (
    "This list is an owned function argument. Therefore, this mutation won't be "
    "visible to the caller. Consider calling `copy()` to obtain a mutable local copy "
    "of this list."
)


class frozenlist(list):  # type: ignore[type-arg]
    """An immutable list subclass.

    Raises a `GuppyComptimeError` for any operation that would mutate the list.
    """

    def append(self, *args: Any, **kwargs: Any) -> None:
        raise GuppyComptimeError(ERROR_MSG)

    def copy(self) -> Any:
        return list(self)

    def clear(self, *args: Any, **kwargs: Any) -> None:
        raise GuppyComptimeError(ERROR_MSG)

    def extend(self, *args: Any, **kwargs: Any) -> None:
        raise GuppyComptimeError(ERROR_MSG)

    def insert(self, *args: Any, **kwargs: Any) -> None:
        raise GuppyComptimeError(ERROR_MSG)

    def pop(self, *args: Any, **kwargs: Any) -> Any:
        raise GuppyComptimeError(ERROR_MSG)

    def remove(self, *args: Any, **kwargs: Any) -> None:
        raise GuppyComptimeError(ERROR_MSG)

    def reverse(self, *args: Any, **kwargs: Any) -> None:
        raise GuppyComptimeError(ERROR_MSG)

    def sort(self, *args: Any, **kwargs: Any) -> None:
        raise GuppyComptimeError(ERROR_MSG)

    def __delitem__(self, *args: Any, **kwargs: Any) -> None:
        raise GuppyComptimeError(ERROR_MSG)

    def __iadd__(self, *args: Any, **kwargs: Any) -> Self:  # type: ignore[misc]
        raise GuppyComptimeError(ERROR_MSG)

    def __imul__(self, *args: Any, **kwargs: Any) -> Self:  # type: ignore[misc]
        raise GuppyComptimeError(ERROR_MSG)

    def __setitem__(self, key: Any, value: Any) -> Any:
        raise GuppyComptimeError(ERROR_MSG)



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.declare
def foo(q1: qubit) -> None: ...


@guppy.declare
def use(q: qubit @owned) -> None: ...


@guppy
def test(q: qubit @owned) -> None:
    use(q)
    foo(q)


test.compile()



from guppylang import guppy

T = guppy.type_var("T")


@guppy
def foo(x: T, y: float) -> T: ...


@guppy
def bar(x: float, y: T) -> T: ...


@guppy.overload(foo, bar)
def overloaded(): ...


@guppy
def main() -> None:
    x: int = overloaded(1.2, 3.4)


main.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit, measure


@guppy.struct
class MyStruct1:
    x: "MyStruct2"


@guppy.struct
class MyStruct2:
    q: qubit


@guppy.declare
def use(s: MyStruct1 @owned) -> None: ...


@guppy
def foo(s: MyStruct1 @owned) -> MyStruct1:
    use(s)
    return s


foo.compile()



from guppylang.decorator import guppy


@guppy.struct
class MyStruct:
    x: int
    y: bool


@guppy
def foo() -> MyStruct:
    return MyStruct(42, False)


@guppy
def bar() -> None:
    foo().x += 1


bar.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy.declare
def bar(q: qubit @owned) -> list[int]:
    ...


@guppy
def foo(qs: list[qubit] @owned) -> list[qubit]:
    return [q for q in qs for x in bar(q)]


foo.compile()



from typing import Callable

from guppylang.decorator import guppy

T = guppy.type_var("T")


@guppy.declare
def foo(f: Callable[[T], T]) -> None:
    ...


@guppy.declare
def bar(x: T) -> T:
    ...


@guppy
def main() -> None:
    foo(bar)


main.compile()



from hugr import tys
from guppylang import guppy
from guppylang_internals.decorator import custom_type


@custom_type(tys.Tuple())
class T:
    pass


@guppy.declare
def foo(x: T) -> None: ...


@guppy.comptime
def test() -> None:
    foo(T)


test.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import array, owned


def test_turns_into_list(run_int_fn):
    @guppy.comptime
    def test(xs: array[int, 10]) -> int:
        assert isinstance(xs, list)
        assert len(xs) == 10

        return sum(xs)

    @guppy
    def main() -> int:
        return test(array(i for i in range(10)))

    run_int_fn(main, sum(range(10)))


def test_accepts_list(run_int_fn):
    @guppy
    def foo(xs: array[int, 10] @ owned) -> int:
        s = 0
        for x in xs:
            s += x
        return s

    @guppy.comptime
    def main() -> int:
        return foo(list(range(10)))

    run_int_fn(main, sum(range(10)))


def test_create(run_int_fn):
    @guppy.comptime
    def main() -> int:
        xs = array(*range(10))
        assert isinstance(xs, list)
        assert xs == list(range(10))
        return xs[-1]

    run_int_fn(main, 9)


def test_mutate(run_int_fn):
    @guppy.comptime
    def test(xs: array[int, 10]) -> None:
        ys = xs
        ys[0] = 100

    @guppy
    def main() -> int:
        xs = array(i for i in range(10))
        test(xs)
        return xs[0]

    run_int_fn(main, 100)


def test_comprehension(validate):
    @guppy.comptime
    def main() -> array[int, 5]:
        xs = array(i for i in range(5))
        assert xs == [0, 1, 2, 3, 4]
        return xs

    validate(main.compile())



"""Compilers building list functions on top of hugr standard operations, that involve
multiple nodes.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import hugr.std.collections.list
from hugr import Wire, ops
from hugr import tys as ht
from hugr.std.collections.list import List, ListVal

from guppylang_internals.definition.custom import CustomCallCompiler
from guppylang_internals.definition.value import CallReturnWires
from guppylang_internals.error import InternalGuppyError
from guppylang_internals.std._internal.compiler.arithmetic import (
    convert_ifromusize,
    convert_itousize,
)
from guppylang_internals.std._internal.compiler.prelude import (
    build_unwrap,
    build_unwrap_left,
    build_unwrap_right,
)
from guppylang_internals.tys.arg import TypeArg

if TYPE_CHECKING:
    from hugr.build.dfg import DfBase


# ------------------------------------------------------
# --------- std.collections operations -----------------
# ------------------------------------------------------


def _instantiate_list_op(
    name: str, elem_type: ht.Type, inp: list[ht.Type], out: list[ht.Type]
) -> ops.ExtOp:
    op_def = hugr.std.collections.list.EXTENSION.get_op(name)
    return ops.ExtOp(
        op_def,
        ht.FunctionType(inp, out),
        [ht.TypeTypeArg(elem_type)],
    )


def list_pop(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `pop` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op(
        "pop", elem_type, [list_type], [list_type, ht.Option(elem_type)]
    )


def list_push(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `push` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op("push", elem_type, [list_type, elem_type], [list_type])


def list_get(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `get` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op(
        "get", elem_type, [list_type, ht.USize()], [ht.Option(elem_type)]
    )


def list_set(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `set` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op(
        "set",
        elem_type,
        [list_type, ht.USize(), elem_type],
        [list_type, ht.Either([elem_type], [elem_type])],
    )


def list_insert(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `insert` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op(
        "insert",
        elem_type,
        [list_type, ht.USize(), elem_type],
        [list_type, ht.Either([elem_type], [ht.Unit])],
    )


def list_length(elem_type: ht.Type) -> ops.ExtOp:
    """Returns a list `length` operation."""
    list_type = List(elem_type)
    return _instantiate_list_op(
        "length", elem_type, [list_type], [list_type, ht.USize()]
    )


# ------------------------------------------------------
# --------- Custom compilers for non-native ops --------
# ------------------------------------------------------


class ListGetitemCompiler(CustomCallCompiler):
    """Compiler for the `list.__getitem__` function."""

    def build_classical_getitem(
        self,
        list_wire: Wire,
        idx: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.__getitem__` for classical lists."""
        idx = self.builder.add_op(convert_itousize(), idx)
        result = self.builder.add_op(list_get(elem_ty), list_wire, idx)
        elem = build_unwrap(self.builder, result, "List index out of bounds")
        return CallReturnWires(regular_returns=[elem], inout_returns=[list_wire])

    def build_linear_getitem(
        self,
        list_wire: Wire,
        idx: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.__getitem__` for linear lists."""
        # Swap out the element at the given index with `None`. The `to_hugr`
        # implementation of the list type ensures that linear element types are turned
        # into optionals.
        elem_opt_ty = ht.Option(elem_ty)
        none = self.builder.add_op(ops.Tag(0, elem_opt_ty))
        idx = self.builder.add_op(convert_itousize(), idx)
        list_wire, result = self.builder.add_op(
            list_set(elem_opt_ty), list_wire, idx, none
        )
        elem_opt = build_unwrap_right(self.builder, result, "List index out of bounds")
        elem = build_unwrap(
            self.builder, elem_opt, "Linear list element has already been used"
        )
        return CallReturnWires(regular_returns=[elem], inout_returns=[list_wire])

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [list_wire, idx] = args
        [elem_ty_arg] = self.type_args
        assert isinstance(elem_ty_arg, TypeArg)
        if elem_ty_arg.ty.linear:
            return self.build_linear_getitem(
                list_wire, idx, elem_ty_arg.ty.to_hugr(self.ctx)
            )
        else:
            return self.build_classical_getitem(
                list_wire, idx, elem_ty_arg.ty.to_hugr(self.ctx)
            )

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


class ListSetitemCompiler(CustomCallCompiler):
    """Compiler for the `list.__setitem__` function."""

    def build_classical_setitem(
        self,
        list_wire: Wire,
        idx: Wire,
        elem: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.__setitem__` for classical lists."""
        idx = self.builder.add_op(convert_itousize(), idx)
        list_wire, result = self.builder.add_op(list_set(elem_ty), list_wire, idx, elem)
        # Unwrap the result, but we don't have to hold onto the returned old value
        build_unwrap_right(self.builder, result, "List index out of bounds")
        return CallReturnWires(regular_returns=[], inout_returns=[list_wire])

    def build_linear_setitem(
        self,
        list_wire: Wire,
        idx: Wire,
        elem: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `array.__setitem__` for linear arrays."""
        # Embed the element into an optional
        elem_opt_ty = ht.Option(elem_ty)
        elem = self.builder.add_op(ops.Some(elem_ty), elem)
        idx = self.builder.add_op(convert_itousize(), idx)
        list_wire, result = self.builder.add_op(
            list_set(elem_opt_ty), list_wire, idx, elem
        )
        old_elem_opt = build_unwrap_right(
            self.builder, result, "List index out of bounds"
        )
        # Check that the old element was `None`
        build_unwrap_left(
            self.builder, old_elem_opt, "Linear list element has not been used"
        )
        return CallReturnWires(regular_returns=[], inout_returns=[list_wire])

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [list_wire, idx, elem] = args
        [elem_ty_arg] = self.type_args
        assert isinstance(elem_ty_arg, TypeArg)
        if elem_ty_arg.ty.linear:
            return self.build_linear_setitem(
                list_wire, idx, elem, elem_ty_arg.ty.to_hugr(self.ctx)
            )
        else:
            return self.build_classical_setitem(
                list_wire, idx, elem, elem_ty_arg.ty.to_hugr(self.ctx)
            )

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


class ListPopCompiler(CustomCallCompiler):
    """Compiler for the `list.pop` function."""

    def build_classical_pop(
        self,
        list_wire: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.pop` for classical lists."""
        list_wire, result = self.builder.add_op(list_pop(elem_ty), list_wire)
        elem = build_unwrap(self.builder, result, "List index out of bounds")
        return CallReturnWires(regular_returns=[elem], inout_returns=[list_wire])

    def build_linear_pop(
        self,
        list_wire: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.pop` for linear lists."""
        elem_opt_ty = ht.Option(elem_ty)
        list_wire, result = self.builder.add_op(list_pop(elem_opt_ty), list_wire)
        elem_opt = build_unwrap(self.builder, result, "List index out of bounds")
        elem = build_unwrap(
            self.builder, elem_opt, "Linear list element has already been used"
        )
        return CallReturnWires(regular_returns=[elem], inout_returns=[list_wire])

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [list_wire] = args
        [elem_ty_arg] = self.type_args
        assert isinstance(elem_ty_arg, TypeArg)
        if elem_ty_arg.ty.linear:
            return self.build_linear_pop(list_wire, elem_ty_arg.ty.to_hugr(self.ctx))
        else:
            return self.build_classical_pop(list_wire, elem_ty_arg.ty.to_hugr(self.ctx))

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


class ListPushCompiler(CustomCallCompiler):
    """Compiler for the `list.push` function."""

    def build_classical_push(
        self,
        list_wire: Wire,
        elem: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.push` for classical lists."""
        list_wire = self.builder.add_op(list_push(elem_ty), list_wire, elem)
        return CallReturnWires(regular_returns=[], inout_returns=[list_wire])

    def build_linear_push(
        self,
        list_wire: Wire,
        elem: Wire,
        elem_ty: ht.Type,
    ) -> CallReturnWires:
        """Lowers a call to `list.push` for linear lists."""
        # Wrap element into an optional
        elem_opt_ty = ht.Option(elem_ty)
        elem_opt = self.builder.add_op(ops.Some(elem_ty), elem)
        list_wire = self.builder.add_op(list_push(elem_opt_ty), list_wire, elem_opt)
        return CallReturnWires(regular_returns=[], inout_returns=[list_wire])

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [list_wire, elem] = args
        [elem_ty_arg] = self.type_args
        assert isinstance(elem_ty_arg, TypeArg)
        if elem_ty_arg.ty.linear:
            return self.build_linear_push(
                list_wire, elem, elem_ty_arg.ty.to_hugr(self.ctx)
            )
        else:
            return self.build_classical_push(
                list_wire, elem, elem_ty_arg.ty.to_hugr(self.ctx)
            )

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


class ListLengthCompiler(CustomCallCompiler):
    """Compiler for the `list.__len__` function."""

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [list_wire] = args
        [elem_ty_arg] = self.type_args
        assert isinstance(elem_ty_arg, TypeArg)
        elem_ty = elem_ty_arg.ty.to_hugr(self.ctx)
        if elem_ty_arg.ty.linear:
            elem_ty = ht.Option(elem_ty)
        list_wire, length = self.builder.add_op(list_length(elem_ty), list_wire)
        length = self.builder.add_op(convert_ifromusize(), length)
        return CallReturnWires(regular_returns=[length], inout_returns=[list_wire])

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


def list_new(
    builder: DfBase[ops.DfParentOp], elem_type: ht.Type, args: list[Wire]
) -> Wire:
    if elem_type.type_bound() == ht.TypeBound.Linear:
        return _list_new_linear(builder, elem_type, args)
    else:
        return _list_new_classical(builder, elem_type, args)


def _list_new_classical(
    builder: DfBase[ops.DfParentOp], elem_type: ht.Type, args: list[Wire]
) -> Wire:
    # This may be simplified in the future with a `new` or `with_capacity` list op
    # See https://github.com/CQCL/hugr/issues/1508
    lst = builder.load(ListVal([], elem_ty=elem_type))
    push_op = list_push(elem_type)
    for elem in args:
        lst = builder.add_op(push_op, lst, elem)
    return lst


def _list_new_linear(
    builder: DfBase[ops.DfParentOp], elem_type: ht.Type, args: list[Wire]
) -> Wire:
    elem_opt_ty = ht.Option(elem_type)
    lst = builder.load(ListVal([], elem_ty=elem_opt_ty))
    push_op = list_push(elem_opt_ty)
    for elem in args:
        elem_opt = builder.add_op(ops.Some(elem_type), elem)
        lst = builder.add_op(push_op, lst, elem_opt)
    return lst



from guppylang.decorator import guppy


def test_factorial(validate):
    @guppy
    def factorial1(x: int) -> int:
        acc = 1
        while x > 0:
            acc *= x
            x -= 1
        return acc

    @guppy
    def factorial2(x: int) -> int:
        if x == 0:
            return 1
        return factorial2(x - 1) * x

    @guppy
    def factorial3(x: int, acc: int) -> int:
        if x == 0:
            return acc
        return factorial3(x - 1, acc * x)

    validate(factorial1.compile(), name="factorial1")
    validate(factorial2.compile(), name="factorial2")
    validate(factorial3.compile(), name="factorial3")


def test_even_odd(validate):
    @guppy
    def is_even(x: int) -> bool:
        if x == 0:
            return True
        return is_odd(x - 1)

    @guppy
    def is_odd(x: int) -> bool:
        if x == 0:
            return False
        return is_even(x - 1)

    validate(is_odd.compile())
    validate(is_even.compile())



import guppylang_internals
from guppylang_internals.engine import CoreMetadataKeys

import guppylang
from guppylang import guppy


def test_metadata():
    @guppy
    def foo() -> None:
        pass

    hugr = foo.compile().modules[0]
    meta = hugr.module_root.metadata
    gen_key = CoreMetadataKeys.GENERATOR.value
    assert (
        meta[gen_key]["name"]
        == f"guppylang (guppylang-internals-v{guppylang_internals.__version__})"
    )
    assert meta[gen_key]["version"] == guppylang.__version__

    used_key = CoreMetadataKeys.USED_EXTENSIONS.value
    used = meta[used_key]
    assert len(used) > 0
    assert all("name" in ext and "version" in ext for ext in used)
    assert all(isinstance(ext["name"], str) for ext in used)
    assert all(isinstance(ext["version"], str) for ext in used)



from typing import Generic

from hugr import tys
from guppylang import guppy
from guppylang_internals.decorator import custom_type


@custom_type(tys.Tuple(), droppable=False)
class MyType:
    pass


X = guppy.const_var("X", "MyType")


@guppy.struct
class Struct(Generic[X]):
    pass


@guppy
def main(_: Struct[X]) -> None:
    pass


main.compile()



from guppylang.decorator import guppy
from guppylang.std.angles import angle
from guppylang.std.builtins import owned, array
from guppylang.std.qsystem.random import make_discrete_distribution, RNG

from guppylang.std.qsystem import MaybeLeaked, measure_leaked
from guppylang.std.qsystem.utils import get_current_shot
from guppylang.std.quantum import qubit, measure_array
from guppylang.std.qsystem.functional import (
    phased_x,
    zz_phase,
    measure_and_reset,
    zz_max,
    reset,
    rz,
    measure,
    qfree,
)


def test_qsystem(validate):  # type: ignore[no-untyped-def]
    """Compile various operations from the qsystem extension."""

    @guppy
    def test(q1: qubit @ owned, q2: qubit @ owned, a1: angle) -> bool:
        shot = get_current_shot()
        q1 = phased_x(q1, a1, a1)
        q1, q2 = zz_phase(q1, q2, a1)
        q1 = rz(q1, a1)
        q1, q2 = zz_max(q1, q2)
        q1, b = measure_and_reset(q1)
        q1 = reset(q1)
        b = measure(q1)
        qfree(q2)
        return b

    validate(test.compile())


def test_qsystem_random(validate):  # type: ignore[no-untyped-def]
    """Compile various operations from the qsystem random extension."""

    @guppy
    def test() -> tuple[int, float, int, int, angle, angle]:
        rng = RNG(42)
        rint = rng.random_int()
        rfloat = rng.random_float()
        rint_bnd = rng.random_int_bounded(100)
        ar = array(qubit() for _ in range(5))
        rng.shuffle(ar)
        _ = measure_array(ar)
        dist = make_discrete_distribution(array(0.0, 1.0, 2.0, 3.0))
        rint_discrete = dist.sample(rng)
        rangle = rng.random_angle()
        rcangle = rng.random_clifford_angle()
        rng.discard()

        return rint, rfloat, rint_bnd, rint_discrete, rangle, rcangle

    validate(test.compile())


def test_measure_leaked(validate):  # type: ignore[no-untyped-def]
    """Compile the measure_leaked operation."""

    @guppy
    def test(q: qubit @ owned) -> bool:
        ml: MaybeLeaked = measure_leaked(q)
        if ml.is_leaked():
            ml.discard()
            return False
        b: bool = ml.to_result().unwrap()
        return b

    validate(test.compile())



from typing import Generic

from guppylang.decorator import guppy

F = guppy.const_var("F", "float")


@guppy.struct
class Struct(Generic[F]):
    pass


@guppy
def main(_: Struct[F]) -> float:
    return F


main.compile()



from guppylang.decorator import guppy

X = guppy.type_var("X")


class Generic:
    """Fake Generic type that doesn't check for type var uniqueness."""
    def __class_getitem__(cls, item):
        return cls


@guppy.struct
class MyStruct(Generic[X, X]):
    x: int


MyStruct.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import array, owned


@guppy.declare
def foo(q: array[int, 3]) -> None: ...


@guppy.declare
def use(q: array[int, 3] @owned) -> None: ...


@guppy
def test(q: array[int, 3]) -> None:
    foo(q)
    use(q)


test.compile()



from guppylang.decorator import guppy


@guppy.struct
class S:
    x: int

    @guppy.declare
    def __iter__(self: "S") -> int: ...


@guppy.comptime
def test() -> None:
    s = S(1)
    s.x = 1.0
    # Even though `S` implements `__iter__`, we can't use it during tracing
    for _ in s:
        pass


test.compile()



from guppylang import guppy
from guppylang_internals.decorator import wasm, wasm_module
from guppylang.std.quantum import qubit


@wasm_module("", 0)
class Foo:
    @wasm
    def foo(self: "Foo", x: qubit) -> qubit: ...

@guppy
def main() -> qubit:
    mod = Foo(0)
    q = mod.foo(qubit())
    mod.discard
    return q

main.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import array, owned


@guppy.declare
def bar(xs: array[int, 1]) -> None: ...


@guppy.comptime
def foo(xs: array[int, 1] @ owned) -> None:
    bar(xs)
    # Remains immutable after an inout call
    del xs[0]


@guppy.comptime
def main() -> None:
    xs = [0]
    foo(xs)
    # Python users would expect xs to be mutated by the call above. However, we cannot
    # provide this semantics, so the mutation in `foo` is rejected by the compiler.


main.compile()



from guppylang.decorator import guppy


@guppy
def f(x: int) -> int:
    return x


@guppy
def g(x: int) -> int:
    return x


@guppy
def main() -> tuple[int, int]:
    return (f, g)(1, 2)


main.compile()



from collections.abc import Callable

from guppylang import guppy
from guppylang.std.builtins import comptime, nat


@guppy
def main(f: Callable[[nat @comptime], None]) -> None:
    pass


main.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import nat, bytecast_nat_to_float, bytecast_float_to_nat


def test_roundtrip(run_int_fn):
    @guppy
    def roundtrip(n: nat) -> nat:
        f = bytecast_nat_to_float(n)
        return bytecast_float_to_nat(f)

    @guppy
    def main() -> nat:
        x = roundtrip(42)
        return x

    run_int_fn(main, expected=42)


def test_roundtrip2(run_float_fn_approx):
    @guppy
    def roundtrip(f: float) -> float:
        n = bytecast_float_to_nat(f)
        f2 = bytecast_nat_to_float(n)
        n2 = bytecast_float_to_nat(f2)
        return bytecast_nat_to_float(n2)

    @guppy
    def main() -> float:
        x = roundtrip(42.0)
        return x

    run_float_fn_approx(main, expected=42.0)



"""
Emulator module for GuppyLang.

This module provides classes for building and executing emulators that can run
Guppy programs. It includes functionality for extracting emulator state,
building and configuring emulator instances, and processing emulation results.
"""

from .builder import EmulatorBuilder
from .instance import EmulatorInstance
from .result import EmulatorResult
from .state import PartialVector, StateVector, TracedState

__all__ = [
    "EmulatorInstance",
    "EmulatorResult",
    "EmulatorBuilder",
    "TracedState",
    "PartialVector",
    "StateVector",
]



from typing import no_type_check

from guppylang import guppy
from guppylang.std.builtins import array, owned
from guppylang.std.debug import state_result
from guppylang.std.quantum import qubit, discard, discard_array, x, h, cx
from tests.util import compile_guppy


def test_basic(validate):
    @compile_guppy
    def main() -> None:
        q1 = qubit()
        q2 = qubit()
        state_result("tag", q1, q2)
        discard(q1)
        discard(q2)

    validate(main)


def test_multi(validate):
    @compile_guppy
    def main() -> None:
        q1 = qubit()
        q2 = qubit()
        state_result("tag1", q1, q2)
        x(q1)
        state_result("tag2", q1)
        discard(q1)
        discard(q2)

    validate(main)


def test_array_access(validate):
    @compile_guppy
    def main() -> None:
        qs = array(qubit() for _ in range(4))
        h(qs[1])
        h(qs[2])
        state_result("a", qs[1], qs[2], qs[3])
        state_result("b", qs[1])
        h(qs[3])

        cx(qs[1], qs[2])
        state_result("c", qs[2], qs[3])
        cx(qs[3], qs[4])
        discard_array(qs)

    validate(main)


def test_struct_access(validate):
    @guppy.struct
    class S:
        q1: qubit
        q2: qubit
        q3: qubit
        q4: qubit

    @guppy
    @no_type_check
    def test(qs: S @ owned) -> None:
        h(qs.q1)
        h(qs.q2)
        state_result("1", qs.q1, qs.q2, qs.q3)
        state_result("2", qs.q1)
        h(qs.q3)

        cx(qs.q1, qs.q2)
        state_result("3", qs.q2, qs.q3)
        cx(qs.q3, qs.q4)

        discard(qs.q1)
        discard(qs.q2)
        discard(qs.q3)
        discard(qs.q4)

    validate(test.compile())


def test_array(validate):
    @compile_guppy
    def main() -> None:
        qs = array(qubit() for _ in range(4))
        h(qs[1])
        h(qs[2])
        cx(qs[0], qs[3])
        state_result("tag", qs)
        discard_array(qs)

    validate(main)


def test_generic_array(validate):
    n = guppy.nat_var("n")

    @guppy
    def main(qs: array[qubit, n]) -> None:
        cx(qs[0], qs[1])
        state_result("tag", qs)

    validate(main.compile())



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy.struct
class MyStruct:
    q: qubit


@guppy
def foo(xs: list[int], s: MyStruct @owned) -> list[MyStruct]:
    return [s for x in xs]


foo.compile()



from dataclasses import dataclass

from hugr import ops
from hugr import tys as ht
from hugr import val as hv

from guppylang_internals.std._internal.compiler.tket_exts import BOOL_EXTENSION

BOOL_DEF = BOOL_EXTENSION.get_type("bool")
OpaqueBool = ht.ExtType(BOOL_DEF)


def read_bool() -> ops.ExtOp:
    return ops.ExtOp(
        BOOL_EXTENSION.get_op("read"),
        ht.FunctionType([OpaqueBool], [ht.Bool]),
    )


def make_opaque() -> ops.ExtOp:
    return ops.ExtOp(
        BOOL_EXTENSION.get_op("make_opaque"),
        ht.FunctionType([ht.Bool], [OpaqueBool]),
    )


def not_op() -> ops.ExtOp:
    return ops.ExtOp(
        BOOL_EXTENSION.get_op("not"),
        ht.FunctionType([OpaqueBool], [OpaqueBool]),
    )


@dataclass
class OpaqueBoolVal(hv.ExtensionValue):
    """Custom value for a boolean."""

    v: bool

    def to_value(self) -> hv.Extension:
        name = "ConstBool"
        payload = self.v
        return hv.Extension(
            name,
            typ=OpaqueBool,
            val=payload,
            extensions=[BOOL_EXTENSION.name],
        )

    def __str__(self) -> str:
        return f"{self.v}"


OPAQUE_TRUE = OpaqueBoolVal(True)
OPAQUE_FALSE = OpaqueBoolVal(False)



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.mem import with_owned
from guppylang.std.quantum import qubit, measure


def test_with_owned(validate):
    @guppy
    def measure_and_reset(q: qubit) -> bool:
        def helper(q: qubit @ owned) -> tuple[bool, qubit]:
            return measure(q), qubit()

        return with_owned(q, helper)

    validate(measure_and_reset.compile())



from guppylang import guppy


@guppy
def foo(x: int) -> None: ...


@guppy
def bar(x: int, y: int) -> None: ...


@guppy.overload(foo, bar)
def overloaded(): ...


@guppy
def main() -> None:
    overloaded(1, 2, 3)


main.compile()



from collections import defaultdict
from enum import Enum
from types import FrameType
from typing import TYPE_CHECKING

import hugr.build.function as hf
import hugr.std.collections.array
import hugr.std.float
import hugr.std.int
import hugr.std.logic
import hugr.std.prelude
from hugr import ops
from hugr.ext import Extension
from hugr.package import ModulePointer, Package

import guppylang_internals
from guppylang_internals.definition.common import (
    CheckableDef,
    CheckedDef,
    CompiledDef,
    DefId,
    ParsableDef,
    ParsedDef,
    RawDef,
)
from guppylang_internals.definition.ty import TypeDef
from guppylang_internals.definition.value import (
    CompiledCallableDef,
    CompiledHugrNodeDef,
)
from guppylang_internals.error import pretty_errors
from guppylang_internals.span import SourceMap
from guppylang_internals.tys.builtin import (
    array_type_def,
    bool_type_def,
    callable_type_def,
    float_type_def,
    frozenarray_type_def,
    int_type_def,
    list_type_def,
    nat_type_def,
    none_type_def,
    option_type_def,
    sized_iter_type_def,
    string_type_def,
    tuple_type_def,
)

if TYPE_CHECKING:
    from guppylang_internals.compiler.core import MonoDefId

BUILTIN_DEFS_LIST: list[RawDef] = [
    callable_type_def,
    tuple_type_def,
    none_type_def,
    bool_type_def,
    nat_type_def,
    int_type_def,
    float_type_def,
    string_type_def,
    list_type_def,
    array_type_def,
    frozenarray_type_def,
    sized_iter_type_def,
    option_type_def,
]

BUILTIN_DEFS = {defn.name: defn for defn in BUILTIN_DEFS_LIST}


class CoreMetadataKeys(Enum):
    """Core HUGR metadata keys used by Guppy."""

    USED_EXTENSIONS = "core.used_extensions"
    GENERATOR = "core.generator"


class DefinitionStore:
    """Storage class holding references to all Guppy definitions created in the current
    interpreter session.

    See `DEF_STORE` for the singleton instance of this class.
    """

    raw_defs: dict[DefId, RawDef]
    impls: defaultdict[DefId, dict[str, DefId]]
    frames: dict[DefId, FrameType]
    sources: SourceMap

    def __init__(self) -> None:
        self.raw_defs = {defn.id: defn for defn in BUILTIN_DEFS_LIST}
        self.impls = defaultdict(dict)
        self.frames = {}
        self.sources = SourceMap()

    def register_def(self, defn: RawDef, frame: FrameType | None) -> None:
        self.raw_defs[defn.id] = defn
        if frame:
            self.frames[defn.id] = frame

    def register_impl(self, ty_id: DefId, name: str, impl_id: DefId) -> None:
        self.impls[ty_id][name] = impl_id
        # Update the frame of the definition to the frame of the defining class
        if impl_id in self.frames:
            frame = self.frames[impl_id].f_back
            if frame:
                self.frames[impl_id] = frame
                # For Python 3.12 generic functions and classes, there is an additional
                # inserted frame for the annotation scope. We can detect this frame by
                # looking for the special ".generic_base" variable in the frame locals
                # that is implicitly inserted by CPython. See
                # - https://docs.python.org/3/reference/executionmodel.html#annotation-scopes
                # - https://docs.python.org/3/reference/compound_stmts.html#generic-functions
                # - https://jellezijlstra.github.io/pep695.html
                if ".generic_base" in frame.f_locals:
                    frame = frame.f_back
                    assert frame is not None
                    self.frames[impl_id] = frame


DEF_STORE: DefinitionStore = DefinitionStore()


class CompilationEngine:
    """Main compiler driver handling checking and compiling of definitions.

    The engine maintains a worklist of definitions that still need to be checked and
    makes sure that all dependencies are compiled.

    See `ENGINE` for the singleton instance of this class.
    """

    parsed: dict[DefId, ParsedDef]
    checked: dict[DefId, CheckedDef]
    compiled: dict["MonoDefId", CompiledDef]
    additional_extensions: list[Extension]

    types_to_check_worklist: dict[DefId, ParsedDef]
    to_check_worklist: dict[DefId, ParsedDef]

    def reset(self) -> None:
        """Resets the compilation cache."""
        self.parsed = {}
        self.checked = {}
        self.compiled = {}
        self.additional_extensions = []
        self.to_check_worklist = {}
        self.types_to_check_worklist = {}

    @pretty_errors
    def register_extension(self, extension: Extension) -> None:
        self.additional_extensions.append(extension)

    @pretty_errors
    def get_parsed(self, id: DefId) -> ParsedDef:
        """Look up the parsed version of a definition by its id.

        Parses the definition if it hasn't been parsed yet. Also makes sure that the
        definition will be checked and compiled later on.
        """
        from guppylang_internals.checker.core import Globals

        if id in self.parsed:
            return self.parsed[id]
        defn = DEF_STORE.raw_defs[id]
        if isinstance(defn, ParsableDef):
            defn = defn.parse(Globals(DEF_STORE.frames[defn.id]), DEF_STORE.sources)
        self.parsed[id] = defn
        if isinstance(defn, TypeDef):
            self.types_to_check_worklist[id] = defn
        else:
            self.to_check_worklist[id] = defn
        return defn

    @pretty_errors
    def get_checked(self, id: DefId) -> CheckedDef:
        """Look up the checked version of a definition by its id.

        Parses and checks the definition if it hasn't been parsed/checked yet. Also
        makes sure that the definition will be compiled to Hugr later on.
        """
        from guppylang_internals.checker.core import Globals

        if id in self.checked:
            return self.checked[id]
        defn = self.get_parsed(id)
        if isinstance(defn, CheckableDef):
            defn = defn.check(Globals(DEF_STORE.frames[defn.id]))
        self.checked[id] = defn

        from guppylang_internals.definition.struct import CheckedStructDef

        if isinstance(defn, CheckedStructDef):
            for method_def in defn.generated_methods():
                DEF_STORE.register_def(method_def, None)
                DEF_STORE.register_impl(defn.id, method_def.name, method_def.id)

        return defn

    @pretty_errors
    def check(self, id: DefId) -> None:
        """Top-level function to kick of checking of a definition.

        This is the main driver behind `guppy.check()`.
        """
        from guppylang_internals.checker.core import Globals

        # Clear previous compilation cache.
        # TODO: In order to maintain results from the previous `check` call we would
        #  need to store and check if any dependencies have changed.
        self.reset()

        defn = DEF_STORE.raw_defs[id]
        self.to_check_worklist = {
            defn.id: (
                defn.parse(Globals(DEF_STORE.frames[defn.id]), DEF_STORE.sources)
                if isinstance(defn, ParsableDef)
                else defn
            )
        }
        while self.types_to_check_worklist or self.to_check_worklist:
            # Types need to be checked first. This is because parsing e.g. a function
            # definition requires instantiating the types in its signature which can
            # only be done if the types have already been checked.
            if self.types_to_check_worklist:
                id, _ = self.types_to_check_worklist.popitem()
            else:
                id, _ = self.to_check_worklist.popitem()
            self.checked[id] = self.get_checked(id)

    @pretty_errors
    def compile(self, id: DefId) -> ModulePointer:
        """Top-level function to kick of Hugr compilation of a definition.

        This is the function that is invoked by `guppy.compile`.
        """
        self.check(id)

        # Prepare Hugr for this module
        graph = hf.Module()
        graph.metadata["name"] = "__main__"  # entrypoint metadata

        # Lower definitions to Hugr
        from guppylang_internals.compiler.core import CompilerContext

        ctx = CompilerContext(graph)
        compiled_def = ctx.compile(self.checked[id])
        self.compiled = ctx.compiled

        if (
            isinstance(compiled_def, CompiledHugrNodeDef)
            and isinstance(compiled_def, CompiledCallableDef)
            and not isinstance(graph.hugr[compiled_def.hugr_node].op, ops.FuncDecl)
        ):
            # if compiling a region set it as the HUGR entrypoint
            # can be loosened after https://github.com/CQCL/hugr/issues/2501 is fixed
            graph.hugr.entrypoint = compiled_def.hugr_node

        # TODO: Currently the list of extensions is manually managed by the user.
        #  We should compute this dynamically from the imported dependencies instead.
        #
        # The hugr prelude and std_extensions are implicit.
        from guppylang_internals.std._internal.compiler.tket_exts import TKET_EXTENSIONS

        extensions = [
            *TKET_EXTENSIONS,
            guppylang_internals.compiler.hugr_extension.EXTENSION,
            *self.additional_extensions,
        ]
        # TODO replace with computed extensions after https://github.com/CQCL/guppylang/issues/550
        all_used_extensions = [
            *extensions,
            hugr.std.prelude.PRELUDE_EXTENSION,
            hugr.std.collections.array.EXTENSION,
            hugr.std.float.FLOAT_OPS_EXTENSION,
            hugr.std.float.FLOAT_TYPES_EXTENSION,
            hugr.std.int.INT_OPS_EXTENSION,
            hugr.std.int.INT_TYPES_EXTENSION,
            hugr.std.logic.EXTENSION,
        ]
        graph.hugr.module_root.metadata[CoreMetadataKeys.USED_EXTENSIONS.value] = [
            {
                "name": ext.name,
                "version": str(ext.version),
            }
            for ext in all_used_extensions
        ]
        graph.hugr.module_root.metadata[CoreMetadataKeys.GENERATOR.value] = {
            "name": "guppylang",
            "version": guppylang_internals.__version__,
        }
        return ModulePointer(Package(modules=[graph.hugr], extensions=extensions), 0)


ENGINE: CompilationEngine = CompilationEngine()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.struct
class MyStruct:
    q: qubit


@guppy
def foo(s: MyStruct @owned) -> MyStruct:
    s.q = qubit()
    return s


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import nat, comptime, array


def test_basic_nat(validate):
    @guppy
    def foo(n: nat @ comptime) -> nat:
        return nat(n + 1)

    @guppy
    def main() -> nat:
        return foo(42)

    validate(main.compile())


def test_basic_int(validate):
    @guppy
    def foo(n: int @ comptime) -> int:
        return n + 1

    @guppy
    def main() -> int:
        return foo(42)

    validate(main.compile())


def test_basic_float(validate):
    @guppy
    def foo(f: float @ comptime) -> float:
        return f + 1.5

    @guppy
    def main() -> float:
        return foo(42.0)

    validate(main.compile())


def test_basic_bool(validate):
    @guppy
    def foo(b: bool @ comptime) -> bool:
        return not b

    @guppy
    def main() -> bool:
        return foo(True)

    validate(main.compile())


def test_multiple(validate):
    @guppy
    def foo(
        a: nat @ comptime, b: nat, c: nat @ comptime, d: nat, e: nat, f: nat @ comptime
    ) -> nat:
        return nat(a + b + c + d + e + f)

    @guppy
    def main() -> nat:
        return foo(1, 2, 3, 4, 5, 6)

    validate(main.compile())


def test_comptime_expr(validate):
    @guppy.declare
    def foo(n: nat @ comptime) -> nat: ...

    @guppy
    def main() -> nat:
        return foo(comptime(42 + 1))

    validate(main.compile())


def test_dependent(validate):
    @guppy
    def foo(n: nat @ comptime, xs: "array[int, n]") -> None:  # noqa: F821
        pass

    @guppy
    def main() -> None:
        foo(0, array())
        foo(1, array(1))
        foo(2, array(1, 2))

    validate(main.compile())


def test_dependent_generic(validate):
    x = guppy.nat_var("x")

    @guppy
    def foo(n: nat @ comptime, xs: "array[int, n]") -> None:  # noqa: F821
        pass

    @guppy
    def main(xs: array[int, x]) -> None:
        foo(x, xs)

    validate(main.compile())


def test_type_apply(validate):
    T = guppy.type_var("T")

    @guppy
    def foo(x: T, n: nat @ comptime) -> None:
        pass

    @guppy
    def main(m: nat @ comptime) -> None:
        foo[int, 43](42, 43)
        f = foo[float, 44]
        f(4.2, 44)
        g = foo[nat, m]
        g(42, m)

    validate(main.compile())



from __future__ import annotations

from typing import TYPE_CHECKING

from guppylang.decorator import custom_guppy_decorator, guppy

if TYPE_CHECKING:
    from hugr.package import Package, PackagePointer

    from guppylang.defs import GuppyFunctionDefinition


@custom_guppy_decorator
def compile_guppy(fn) -> Package:
    """A decorator that combines @guppy with HUGR compilation."""
    defn: GuppyFunctionDefinition = guppy(fn)
    return defn.compile()


def dump_llvm(package: PackagePointer):
    try:
        from selene_hugr_qis_compiler import compile_to_llvm_ir

        llvm_module = compile_to_llvm_ir(package.package.to_bytes())
        print(llvm_module)  # noqa: T201

    except ImportError:
        pass



from typing import Generic

from guppylang.decorator import guppy

X = guppy.const_var("X", "float[int]")


@guppy.struct
class Struct(Generic[X]):
    pass


@guppy
def main(_: Struct[X]) -> None:
    pass


main.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit, measure


@guppy.struct
class MyStruct1:
    q: qubit
    x: "MyStruct2"


@guppy.struct
class MyStruct2:
    q: qubit


@guppy
def foo(s: MyStruct1 @owned) -> MyStruct1:
    measure(s.q)
    s.q = s.x.q
    return s


foo.compile()



import itertools
from abc import ABC
from collections import defaultdict
from collections.abc import Callable, Iterator, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, ClassVar, cast

import tket_exts
from hugr import Hugr, Node, Wire, ops
from hugr import tys as ht
from hugr.build import function as hf
from hugr.build.dfg import DP, DefinitionBuilder, DfBase
from hugr.hugr.base import OpVarCov
from hugr.hugr.node_port import ToNode
from hugr.std import PRELUDE
from typing_extensions import assert_never

from guppylang_internals.checker.core import (
    FieldAccess,
    Globals,
    Place,
    PlaceId,
    TupleAccess,
    Variable,
)
from guppylang_internals.definition.common import (
    CheckedDef,
    CompilableDef,
    CompiledDef,
    DefId,
    MonomorphizableDef,
)
from guppylang_internals.definition.ty import TypeDef
from guppylang_internals.definition.value import CompiledCallableDef
from guppylang_internals.diagnostic import Error
from guppylang_internals.engine import ENGINE
from guppylang_internals.error import GuppyError, InternalGuppyError
from guppylang_internals.tys.arg import ConstArg, TypeArg
from guppylang_internals.tys.builtin import nat_type
from guppylang_internals.tys.common import ToHugrContext
from guppylang_internals.tys.const import BoundConstVar, ConstValue
from guppylang_internals.tys.param import ConstParam, Parameter, TypeParam
from guppylang_internals.tys.subst import Inst
from guppylang_internals.tys.ty import (
    BoundTypeVar,
    NumericType,
    StructType,
    TupleType,
    Type,
)

if TYPE_CHECKING:
    from guppylang_internals.tys.arg import Argument

CompiledLocals = dict[PlaceId, Wire]

#: Partial instantiation of generic type parameters for monomorphization.
#:
#: When compiling polymorphic definitions to Hugr, some of their generic parameters will
#: need to be monomorphized (e.g. to support language features that cannot be encoded
#: in Hugr, see `requires_monomorphization` for details). However, other generic
#: parameters can be faithfully captured in Hugr. Thus, we want to perform a *partial*
#: monomorphization when compiling to Hugr.
#:
#: A `mono_args: PartiallyMonomorphizedArgs` represents such a partial monomorphization
#: of generic parameters as a sequence of `Argument | None`. Concretely,
#: `mono_args[i] == None` means that the argument with index `i` is left as a
#: polymorphic Hugr type parameter, i.e. it is not monomorphised by Guppy. Otherwise,
#: `mono_args[i]` specifies the monomorphic instantiation  for parameter `i`. For
#: example, a function with 3 generic parameters that should all be retained in Hugr
#: will have mono_args = `(None, None, None)`.
#:
#: Finally, note that this sequence is required to be a tuple to ensure hashability.
PartiallyMonomorphizedArgs = tuple["Argument | None", ...]


@dataclass(frozen=True)
class GlobalConstId:
    id: int
    base_name: str

    _fresh_ids = itertools.count()

    @staticmethod
    def fresh(base_name: str) -> "GlobalConstId":
        return GlobalConstId(next(GlobalConstId._fresh_ids), base_name)

    @property
    def name(self) -> str:
        return f"{self.base_name}.{self.id}"


#: Unique identifier for a partially monomorphized definition.
#:
#: If the `DefId` corresponds to a `MonomorphizableDef`, then the second tuple entry
#: should hold the corresponding partial monomorphization. Otherwise, the second entry
#: should be `None`.
#:
#: Note the following subtlety: Functions are instances of `MonomorphizableDef`, even if
#: they aren't actually generic! This means that for non-generic function definitions we
#: will have an empty tuple `()` as `PartiallyMonomorphizedArgs`, but never `None`.
#: `None` only shows up for kinds of definitions that are never monomorphized (e.g.
#: definitions of constants).
MonoDefId = tuple[DefId, PartiallyMonomorphizedArgs | None]

#: Unique identifier for global Hugr constants and functions, the latter with optional
#: monomorphized type arguments
MonoGlobalConstId = tuple[GlobalConstId, PartiallyMonomorphizedArgs | None]


@dataclass(frozen=True)
class EntryMonomorphizeError(Error):
    title: ClassVar[str] = "Invalid entry point"
    span_label: ClassVar[str] = (
        "Function `{name}` is not a valid compilation entry point since the value of "
        "generic paramater `{param}` is not known"
    )
    name: str
    param: Parameter


class CompilerContext(ToHugrContext):
    """Compilation context containing all available definitions.

    Maintains a `worklist` of definitions which have been used by other compiled code
    (i.e. `compile_outer` has been called) but have not yet been compiled/lowered
    themselves (i.e. `compile_inner` has not yet been called).
    """

    module: DefinitionBuilder[ops.Module]

    #: The definitions compiled so far. For `MonomorphizableDef`s, their id can occur
    #: multiple times here with respectively different partial monomorphizations. See
    #: `MonoDefId` and `PartiallyMonomorphizedArgs` for details.
    compiled: dict[MonoDefId, CompiledDef]

    # use dict over set for deterministic iteration order
    worklist: dict[MonoDefId, None]

    global_funcs: dict[MonoGlobalConstId, hf.Function]

    #: Partial instantiation for some of the type parameters of the current function for
    #: the purpose of monomorphization
    current_mono_args: PartiallyMonomorphizedArgs | None

    checked_globals: Globals

    def __init__(
        self,
        module: DefinitionBuilder[ops.Module],
    ) -> None:
        self.module = module
        self.worklist = {}
        self.compiled = {}
        self.global_funcs = {}
        self.checked_globals = Globals(None)
        self.current_mono_args = None

    @contextmanager
    def set_monomorphized_args(
        self, mono_args: PartiallyMonomorphizedArgs | None
    ) -> Iterator[None]:
        """Context manager to set the partial monomorphization for the function that is
        compiled currently being compiled.
        """
        old = self.current_mono_args
        self.current_mono_args = mono_args
        yield
        self.current_mono_args = old

    def build_compiled_def(
        self, def_id: DefId, type_args: Inst
    ) -> tuple[CompiledDef, Inst]:
        """Returns the compiled definitions corresponding to the given ID, along with
        the remaining type args that have not been monomorphized away.

        Might mutate the current Hugr if this definition has never been compiled before.
        """
        # TODO: The check below is a hack to support nested function definitions. We
        #  forgot to insert frames for nested functions into the DEF_STORE, which would
        #  make the call to `ENGINE.get_checked` below fail. For now, let's just short-
        #  cut if the function doesn't take any generic params (as is the case for all
        #  nested functions).
        #  See https://github.com/CQCL/guppylang/issues/1032
        if (def_id, ()) in self.compiled:
            assert type_args == []
            return self.compiled[def_id, ()], type_args

        mono_args: PartiallyMonomorphizedArgs | None = None
        rem_args = type_args
        compile_outer: Callable[[], CompiledDef]
        match ENGINE.get_checked(def_id):
            case MonomorphizableDef(params=params) as monomorphizable:
                mono_args, rem_args = partially_monomorphize_args(
                    params, type_args, self
                )
                compile_outer = lambda: monomorphizable.monomorphize(  # noqa: E731 (assign-lambda)
                    self.module, mono_args, self
                )
            case CompilableDef() as compilable:
                compile_outer = lambda: compilable.compile_outer(self.module, self)  # noqa: E731
            case CompiledDef() as compiled_defn:
                compile_outer = lambda: compiled_defn  # noqa: E731
            case defn:
                compile_outer = assert_never(defn)
        if (def_id, mono_args) not in self.compiled:
            self.compiled[def_id, mono_args] = compile_outer()
            self.worklist[def_id, mono_args] = None
        return self.compiled[def_id, mono_args], rem_args

    def compile(self, defn: CheckedDef) -> CompiledDef:
        """Compiles the given definition and all of its dependencies into the current
        Hugr."""
        # Check and compile the entry point
        entry_mono_args: PartiallyMonomorphizedArgs | None = None
        entry_compiled: CompiledDef
        match ENGINE.get_checked(defn.id):
            case MonomorphizableDef(params=params) as defn:
                # Entry point is not allowed to require monomorphization
                for param in params:
                    if requires_monomorphization(param):
                        err = EntryMonomorphizeError(defn.defined_at, defn.name, param)
                        raise GuppyError(err)
                # Thus, the partial monomorphization for the entry point is always empty
                entry_mono_args = tuple(None for _ in params)
                entry_compiled = defn.monomorphize(self.module, entry_mono_args, self)
            case CompilableDef() as defn:
                entry_compiled = defn.compile_outer(self.module, self)
            case CompiledDef() as defn:
                entry_compiled = defn
            case defn:
                entry_compiled = assert_never(defn)

        self.compiled[defn.id, entry_mono_args] = entry_compiled
        self.worklist[defn.id, entry_mono_args] = None

        # Compile definition bodies
        while self.worklist:
            next_id, mono_args = self.worklist.popitem()[0]
            next_def = self.compiled[next_id, mono_args]
            with track_hugr_side_effects(), self.set_monomorphized_args(mono_args):
                next_def.compile_inner(self)

        return entry_compiled

    def build_compiled_instance_func(
        self,
        ty: Type | TypeDef,
        name: str,
        type_args: Inst,
    ) -> tuple[CompiledCallableDef, Inst] | None:
        """Returns s compiled instance method along with its remaining type args that
        have not been monomorphized away, or `None` if the type doesn't have a matching
        method.

        Compiles the definition and all of its dependencies into the current Hugr.
        """
        from guppylang_internals.engine import ENGINE

        parsed_func = self.checked_globals.get_instance_func(ty, name)
        if parsed_func is None:
            return None
        checked_func = ENGINE.get_checked(parsed_func.id)
        compiled_func, rem_args = self.build_compiled_def(checked_func.id, type_args)
        assert isinstance(compiled_func, CompiledCallableDef)
        return compiled_func, rem_args

    def declare_global_func(
        self,
        const_id: GlobalConstId,
        func_ty: ht.PolyFuncType,
        mono_args: PartiallyMonomorphizedArgs | None = None,
    ) -> tuple[hf.Function, bool]:
        """
        Creates a function builder for a global function if it doesn't already exist,
        else returns the existing one.
        """
        if (const_id, mono_args) in self.global_funcs:
            return self.global_funcs[const_id, mono_args], True
        func = self.module.module_root_builder().define_function(
            name=const_id.name,
            input_types=func_ty.body.input,
            output_types=func_ty.body.output,
            type_params=func_ty.params,
        )
        self.global_funcs[const_id, mono_args] = func
        return func, False

    def type_var_to_hugr(self, var: BoundTypeVar) -> ht.Type:
        """Compiles a bound Guppy type variable into a Hugr type.

        Takes care of performing partial monomorphization as specified in the current
        context.
        """
        if self.current_mono_args is None:
            # If we're not inside a monomorphized context, just return the Hugr
            # variable with the same de Bruijn index
            return ht.Variable(var.idx, var.hugr_bound)

        match self.current_mono_args[var.idx]:
            # Either we have a decided to monomorphize the corresponding parameter...
            case TypeArg(ty=ty):
                return ty.to_hugr(self)
            # ... or we're still want to be generic in Hugr
            case None:
                # But in that case we'll have to down-shift the de Bruijn index to
                # account for earlier params that have been monomorphized away
                hugr_idx = compile_variable_idx(var.idx, self.current_mono_args)
                return ht.Variable(hugr_idx, var.hugr_bound)
            case _:
                raise InternalGuppyError("Invalid monomorphization")

    def const_var_to_hugr(self, var: BoundConstVar) -> ht.TypeArg:
        """Compiles a bound Guppy constant variable into a Hugr type argument.

        Takes care of performing partial monomorphization as specified in the current
        context.
        """
        if var.ty != nat_type():
            raise InternalGuppyError(
                "Tried to convert non-nat const type argument to Hugr. This should "
                "have been monomprhized away."
            )
        param = ht.BoundedNatParam(upper_bound=None)

        if self.current_mono_args is None:
            # If we're not inside a monomorphized context, just return the Hugr
            # variable with the same de Bruijn index
            return ht.VariableArg(var.idx, param)

        match self.current_mono_args[var.idx]:
            # Either we have a decided to monomorphize the corresponding parameter...
            case ConstArg(const=ConstValue(value=int(v))):
                return ht.BoundedNatArg(n=v)
            # ... or we're still want to be generic in Hugr
            case None:
                # But in that case we'll have to down-shift the de Bruijn index to
                # account for earlier params that have been monomorphized away
                hugr_idx = compile_variable_idx(var.idx, self.current_mono_args)
                return ht.VariableArg(hugr_idx, param)
            case _:
                raise InternalGuppyError("Invalid monomorphization")


@dataclass
class DFContainer:
    """A dataflow graph under construction.

    This class is passed through the entire compilation pipeline and stores a builder
    for the dataflow child-graph currently being constructed as well as all live local
    variables. Note that the variable map is mutated in-place and always reflects the
    current compilation state.
    """

    builder: DfBase[ops.DfParentOp]
    ctx: CompilerContext
    locals: CompiledLocals = field(default_factory=dict)

    def __init__(
        self,
        builder: DfBase[DP],
        ctx: CompilerContext,
        locals: CompiledLocals | None = None,
    ) -> None:
        generic_builder = cast(DfBase[ops.DfParentOp], builder)
        if locals is None:
            locals = {}
        self.builder = generic_builder
        self.ctx = ctx
        self.locals = locals

    def __getitem__(self, place: Place) -> Wire:
        """Constructs a wire for a local place in this DFG.

        Note that this mutates the Hugr since we might need to pack or unpack some
        tuples to obtain a port for places that involve struct fields.
        """
        # First check, if we already have a wire for this place
        if place.id in self.locals:
            return self.locals[place.id]
        # Otherwise, our only hope is that it's a struct or tuple value that we can
        # rebuild by packing the wires of its constituting fields
        elif isinstance(place.ty, StructType):
            children: list[Place] = [
                FieldAccess(place, field, None) for field in place.ty.fields
            ]
        elif isinstance(place.ty, TupleType):
            children = [
                TupleAccess(place, elem, idx, None)
                for idx, elem in enumerate(place.ty.element_types)
            ]
        else:
            raise InternalGuppyError(f"Couldn't obtain a port for `{place}`")
        child_types = [child.ty.to_hugr(self.ctx) for child in children]
        child_wires = [self[child] for child in children]
        wire = self.builder.add_op(ops.MakeTuple(child_types), *child_wires)[0]
        for child in children:
            if child.ty.linear:
                self.locals.pop(child.id)
        self.locals[place.id] = wire
        return wire

    def __setitem__(self, place: Place, port: Wire) -> None:
        # When assigning a struct value, we immediately unpack it recursively and only
        # store the leaf wires.
        is_return = isinstance(place, Variable) and is_return_var(place.name)
        if isinstance(place.ty, StructType) and not is_return:
            unpack = self.builder.add_op(
                ops.UnpackTuple([t.ty.to_hugr(self.ctx) for t in place.ty.fields]), port
            )
            for field, field_port in zip(place.ty.fields, unpack, strict=True):
                self[FieldAccess(place, field, None)] = field_port
            # If we had a previous wire assigned to this place, we need forget about it.
            # Otherwise, we might use this old value when looking up the place later
            self.locals.pop(place.id, None)
        # Same for tuples.
        elif isinstance(place.ty, TupleType) and not is_return:
            hugr_elem_tys = [ty.to_hugr(self.ctx) for ty in place.ty.element_types]
            unpack = self.builder.add_op(ops.UnpackTuple(hugr_elem_tys), port)
            for idx, (elem, elem_port) in enumerate(
                zip(place.ty.element_types, unpack, strict=True)
            ):
                self[TupleAccess(place, elem, idx, None)] = elem_port
            self.locals.pop(place.id, None)
        else:
            self.locals[place.id] = port

    def __contains__(self, place: Place) -> bool:
        return place.id in self.locals

    def __copy__(self) -> "DFContainer":
        # Make a copy of the var map so that mutating the copy doesn't
        # mutate our variable mapping
        return DFContainer(self.builder, self.ctx, self.locals.copy())


class CompilerBase(ABC):
    """Base class for the Guppy compiler."""

    ctx: CompilerContext

    def __init__(self, ctx: CompilerContext) -> None:
        self.ctx = ctx


def return_var(n: int) -> str:
    """Name of the dummy variable for the n-th return value of a function.

    During compilation, we treat return statements like assignments of dummy variables.
    For example, the statement `return e0, e1, e2` is treated like `%ret0 = e0 ; %ret1 =
    e1 ; %ret2 = e2`. This way, we can reuse our existing mechanism for passing of live
    variables between basic blocks."""
    return f"%ret{n}"


def is_return_var(x: str) -> bool:
    """Checks whether the given name is a dummy return variable."""
    return x.startswith("%ret")


def requires_monomorphization(param: Parameter) -> bool:
    """Checks if a type parameter must be monomorphized before compiling to Hugr.

    This is required for some Guppy language features that cannot be encoded in Hugr
    yet. Currently, this only applies to non-nat const parameters.
    """
    match param:
        case TypeParam():
            return False
        case ConstParam(ty=ty):
            return ty != NumericType(NumericType.Kind.Nat)
        case x:
            return assert_never(x)


def partially_monomorphize_args(
    params: Sequence[Parameter],
    args: Inst,
    ctx: CompilerContext,
) -> tuple[PartiallyMonomorphizedArgs, Inst]:
    """
    Given an instantiation of all type parameters, extracts the ones that will need to
    be monomorphized.

    Also takes care of normalising bound variables w.r.t. the current monomorphization.
    """
    mono_args: list[Argument | None] = []
    rem_args = []
    for param, arg in zip(params, args, strict=True):
        if requires_monomorphization(param):
            # The constant could still refer to a bound variable, so we need to
            # instantiate it w.r.t. to the current monomorphization
            match arg:
                case ConstArg(const=BoundConstVar(idx=idx)):
                    assert ctx.current_mono_args is not None
                    inst = ctx.current_mono_args[idx]
                    assert inst is not None
                    mono_args.append(inst)
                case TypeArg():
                    # TODO: Once we also have type args that require monomorphization,
                    #  we'll need to downshift de Bruijn indices here as well
                    raise NotImplementedError
                case arg:
                    mono_args.append(arg)
        else:
            mono_args.append(None)
            rem_args.append(arg)
    return tuple(mono_args), rem_args


def compile_variable_idx(idx: int, mono_args: PartiallyMonomorphizedArgs) -> int:
    """Returns the Hugr index for a variable.

    Takes care of shifting down Guppy's indices to account for the current partial
    monomorphization. This avoids gaps in Hugr indices due to the fact that not all
    Guppy parameters are lowered to Hugr (some are monomorphized away).
    """
    assert mono_args[idx] is None, "Should not compile monomorphized index"
    return sum(1 for arg in mono_args[:idx] if arg is None)


QUANTUM_EXTENSION = tket_exts.quantum()
RESULT_EXTENSION = tket_exts.result()
DEBUG_EXTENSION = tket_exts.debug()

#: List of extension ops that have side-effects, identified by their qualified name
EXTENSION_OPS_WITH_SIDE_EFFECTS: list[str] = [
    # Results should be order w.r.t. each other but also w.r.t. panics
    *(op_def.qualified_name() for op_def in RESULT_EXTENSION.operations.values()),
    PRELUDE.get_op("panic").qualified_name(),
    PRELUDE.get_op("exit").qualified_name(),
    DEBUG_EXTENSION.get_op("StateResult").qualified_name(),
    # Qubit allocation and deallocation have the side-effect of changing the number of
    # available free qubits
    QUANTUM_EXTENSION.get_op("QAlloc").qualified_name(),
    QUANTUM_EXTENSION.get_op("QFree").qualified_name(),
    QUANTUM_EXTENSION.get_op("MeasureFree").qualified_name(),
]


def may_have_side_effect(op: ops.Op) -> bool:
    """Checks whether an operation could have a side-effect.

    We need to insert implicit state order edges between these kinds of nodes to ensure
    they are executed in the correct order, even if there is no data dependency.
    """
    match op:
        case ops.ExtOp() as ext_op:
            return ext_op.op_def().qualified_name() in EXTENSION_OPS_WITH_SIDE_EFFECTS
        case ops.Custom(op_name=op_name, extension=extension):
            qualified_name = f"{extension}.{op_name}" if extension else op_name
            return qualified_name in EXTENSION_OPS_WITH_SIDE_EFFECTS
        case ops.Call() | ops.CallIndirect():
            # Conservative choice is to assume that all calls could have side effects.
            # In the future we could inspect the call graph to figure out a more
            # precise answer
            return True
        case _:
            return False


@contextmanager
def track_hugr_side_effects() -> Iterator[None]:
    """Initialises the tracking of nodes with side-effects during Hugr building.

    Ensures that state-order edges are implicitly inserted between side-effectful nodes
    to ensure they are executed in the order they are added.
    """
    # Remember original `Hugr.add_node` method that is monkey-patched below.
    hugr_add_node = Hugr.add_node
    # Last node with potential side effects for each dataflow parent
    prev_node_with_side_effect: defaultdict[Node, Node | None] = defaultdict(
        lambda: None
    )

    def hugr_add_node_with_order(
        self: Hugr[OpVarCov],
        op: ops.Op,
        parent: ToNode | None = None,
        num_outs: int | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Node:
        """Monkey-patched version of `Hugr.add_node` that takes care of implicitly
        inserting state order edges between operations that could have side-effects.
        """
        new_node = hugr_add_node(self, op, parent, num_outs, metadata)
        if may_have_side_effect(op):
            handle_side_effect(new_node, self)
        return new_node

    def handle_side_effect(node: Node, hugr: Hugr[OpVarCov]) -> None:
        """Performs the actual order-edge insertion, assuming that `node` has a side-
        effect."""
        parent = hugr[node].parent
        if parent is not None:
            if prev := prev_node_with_side_effect[parent]:
                hugr.add_order_link(prev, node)
            else:
                # If this is the first side-effectful op in this DFG, make a recursive
                # call with the parent since the parent is also considered side-
                # effectful now. We shouldn't walk up through function definitions
                # or basic blocks though
                if not isinstance(hugr[parent].op, ops.FuncDefn | ops.DataflowBlock):
                    handle_side_effect(parent, hugr)
            prev_node_with_side_effect[parent] = node

    # Monkey-patch the `add_node` method
    Hugr.add_node = hugr_add_node_with_order  # type: ignore[method-assign]
    try:
        yield
    finally:
        Hugr.add_node = hugr_add_node  # type: ignore[method-assign]



from collections.abc import Iterator
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from guppylang_internals.ast_util import AstNode
from guppylang_internals.checker.core import Globals
from guppylang_internals.compiler.core import CompilerContext, DFContainer
from guppylang_internals.error import InternalGuppyError

if TYPE_CHECKING:
    from guppylang_internals.tracing.object import GuppyObject, GuppyObjectId


@dataclass
class TracingState:
    """Internal state that is used during the tracing phase of comptime functions."""

    #: Reference to the global compilation context.
    ctx: CompilerContext

    #: The current dataflow graph under construction.
    dfg: DFContainer

    #: An AST node capturing the code block that is currently being traced
    node: AstNode

    #: Set of all allocated undroppable GuppyObjects where the `used` flag is not set,
    #: indexed by their id. This is used to detect linearity violations.
    unused_undroppable_objs: "dict[GuppyObjectId, GuppyObject]" = field(
        default_factory=dict
    )

    @property
    def globals(self) -> Globals:
        return self.ctx.checked_globals


_STATE: ContextVar[TracingState | None] = ContextVar("_STATE", default=None)


def reset_state() -> None:
    """Resets the tracing state to be undefined."""
    _STATE.set(None)


def tracing_active() -> bool:
    """Checks if the tracing mode is currently active."""
    return _STATE.get() is not None


def get_tracing_state() -> TracingState:
    """Returns the current tracing state.

    Raises an `InternalGuppyError` if the tracing mode is currently not active.
    """
    state = _STATE.get()
    if state is None:
        raise InternalGuppyError("Guppy tracing mode is not active")
    return state


@contextmanager
def set_tracing_state(state: TracingState) -> Iterator[None]:
    """Context manager to update tracing state for the duration of a code block."""
    token = _STATE.set(state)
    yield
    _STATE.reset(token)



from typing import Generic, TypeVar

from guppylang.decorator import guppy


X = TypeVar("X")  # This is a Python type variable, not a Guppy one!


@guppy.struct
class MyStruct(Generic[X]):
    x: int


MyStruct.compile()



import ast
from dataclasses import dataclass
from typing import ClassVar, cast

from guppylang_internals.ast_util import with_loc
from guppylang_internals.checker.errors.generic import ExpectedError
from guppylang_internals.checker.errors.type_errors import WrongNumberOfArgsError
from guppylang_internals.checker.expr_checker import (
    ExprChecker,
    ExprSynthesizer,
    synthesize_call,
)
from guppylang_internals.definition.custom import CustomCallChecker
from guppylang_internals.definition.ty import TypeDef
from guppylang_internals.diagnostic import Error
from guppylang_internals.error import GuppyTypeError
from guppylang_internals.nodes import StateResultExpr
from guppylang_internals.std._internal.checker import TAG_MAX_LEN, TooLongError
from guppylang_internals.tys.builtin import (
    get_array_length,
    get_element_type,
    is_array_type,
    string_type,
)
from guppylang_internals.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    Type,
)


class StateResultChecker(CustomCallChecker):
    """Call checker for the `state_result` function."""

    @dataclass(frozen=True)
    class MissingQubitsError(Error):
        title: ClassVar[str] = "Missing qubit inputs"
        span_label: ClassVar[str] = (
            "Qubits whose state should be reported must be passed explicitly"
        )

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        tag, _ = ExprChecker(self.ctx).check(args[0], string_type())
        if not isinstance(tag, ast.Constant) or not isinstance(tag.value, str):
            raise GuppyTypeError(ExpectedError(tag, "a string literal"))
        if len(tag.value.encode("utf-8")) > TAG_MAX_LEN:
            err: Error = TooLongError(tag)
            err.add_sub_diagnostic(TooLongError.Hint(None))
            raise GuppyTypeError(err)
        syn_args: list[ast.expr] = [tag]

        if len(args) < 2:
            raise GuppyTypeError(self.MissingQubitsError(self.node))

        from guppylang.defs import GuppyDefinition
        from guppylang.std.quantum import qubit

        assert isinstance(qubit, GuppyDefinition)
        qubit_ty = cast(TypeDef, qubit.wrapped).check_instantiate([])

        array_len = None
        arg, ty = ExprSynthesizer(self.ctx).synthesize(args[1])
        if is_array_type(ty):
            if len(args) > 2:
                err = WrongNumberOfArgsError(args[2], 2, len(args))
                raise GuppyTypeError(err)
            element_ty = get_element_type(ty)
            if not element_ty == qubit_ty:
                raise GuppyTypeError(ExpectedError(arg, "an array of qubits"))
            syn_args.append(arg)
            func_ty = FunctionType(
                [
                    FuncInput(string_type(), InputFlags.NoFlags),
                    FuncInput(ty, InputFlags.Inout),
                ],
                NoneType(),
            )
            array_len = get_array_length(ty)
        else:
            for arg in args[1:]:
                qbt, _ = ExprChecker(self.ctx).check(arg, qubit_ty)
                syn_args.append(qbt)
            func_ty = FunctionType(
                [FuncInput(string_type(), InputFlags.NoFlags)]
                + [FuncInput(qubit_ty, InputFlags.Inout)] * len(args[1:]),
                NoneType(),
            )
        args, ret_ty, inst = synthesize_call(func_ty, syn_args, self.node, self.ctx)
        assert len(inst) == 0, "func_ty is not generic"
        node = StateResultExpr(
            tag=tag.value, args=args, func_ty=func_ty, array_len=array_len
        )
        return with_loc(self.node, node), ret_ty



from collections.abc import Callable

from guppylang.decorator import guppy


@guppy
def foo(x: int) -> int:
    def bar(f: Callable[[], int]) -> int:
        return f()

    return bar(foo)


foo.compile()



from abc import ABC
from collections.abc import Sequence

from hugr import Wire, ops
from hugr import tys as ht

from guppylang_internals.definition.custom import (
    CustomCallCompiler,
    CustomInoutCallCompiler,
)
from guppylang_internals.definition.value import CallReturnWires
from guppylang_internals.error import InternalGuppyError
from guppylang_internals.std._internal.compiler.prelude import (
    build_unwrap_left,
    build_unwrap_right,
)
from guppylang_internals.std._internal.compiler.tket_bool import (
    OPAQUE_FALSE,
    OPAQUE_TRUE,
)
from guppylang_internals.tys.arg import Argument, TypeArg
from guppylang_internals.tys.common import ToHugrContext
from guppylang_internals.tys.ty import type_to_row


def either_to_hugr(type_args: Sequence[Argument], ctx: ToHugrContext) -> ht.Either:
    match type_args:
        case [TypeArg(left_ty), TypeArg(right_ty)]:
            left_tys = [ty.to_hugr(ctx) for ty in type_to_row(left_ty)]
            right_tys = [ty.to_hugr(ctx) for ty in type_to_row(right_ty)]
            return ht.Either(left_tys, right_tys)
        case _:
            raise InternalGuppyError("Invalid type args for Either type")


class EitherCompiler(CustomInoutCallCompiler, ABC):
    """Abstract base class for compilers for `Either` methods."""

    @property
    def left_tys(self) -> list[ht.Type]:
        match self.type_args:
            case [TypeArg(left_ty), TypeArg()]:
                return [ty.to_hugr(self.ctx) for ty in type_to_row(left_ty)]
            case _:
                raise InternalGuppyError("Invalid type args for Either op")

    @property
    def right_tys(self) -> list[ht.Type]:
        match self.type_args:
            case [TypeArg(), TypeArg(right_ty)]:
                return [ty.to_hugr(self.ctx) for ty in type_to_row(right_ty)]
            case _:
                raise InternalGuppyError("Invalid type args for Either op")

    @property
    def either_ty(self) -> ht.Either:
        return either_to_hugr(self.type_args, self.ctx)


class EitherConstructor(EitherCompiler, CustomCallCompiler):
    """Compiler for the `Option` constructors `nothing` and `some`."""

    def __init__(self, tag: int):
        self.tag = tag

    def compile(self, args: list[Wire]) -> list[Wire]:
        ty = self.either_ty
        if self.tag == 1:
            # In the `right` case, the type args are swapped around since `R` occurs
            # first in the signature :(
            ty.variant_rows = [ty.variant_rows[1], ty.variant_rows[0]]
        return [self.builder.add_op(ops.Tag(self.tag, ty), *args)]


class EitherTestCompiler(EitherCompiler):
    """Compiler for the `Either.is_left` and `Either.is_right` methods."""

    def __init__(self, tag: int):
        self.tag = tag

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [either] = args
        cond = self.builder.add_conditional(either)
        for i in [0, 1]:
            with cond.add_case(i) as case:
                val = OPAQUE_TRUE if i == self.tag else OPAQUE_FALSE
                either = case.add_op(ops.Tag(i, self.either_ty), *case.inputs())
                case.set_outputs(case.load(val), either)
        [res, either] = cond.outputs()
        return CallReturnWires(regular_returns=[res], inout_returns=[either])


class EitherToOptionCompiler(EitherCompiler, CustomCallCompiler):
    """Compiler for the `Either.left` and `Either.right` methods."""

    def __init__(self, tag: int):
        self.tag = tag

    def compile(self, args: list[Wire]) -> list[Wire]:
        [either] = args
        cond = self.builder.add_conditional(either)
        target_tys = self.left_tys if self.tag == 0 else self.right_tys
        for i in [0, 1]:
            with cond.add_case(i) as case:
                if i == self.tag:
                    out = case.add_op(
                        ops.Tag(1, ht.Option(*target_tys)), *case.inputs()
                    )
                else:
                    out = case.add_op(ops.Tag(0, ht.Option(*target_tys)))
                case.set_outputs(out)
        return list(cond.outputs())


class EitherUnwrapCompiler(EitherCompiler, CustomCallCompiler):
    """Compiler for the `Either.unwrap_left` and `Either.unwrap_right` methods."""

    def __init__(self, tag: int):
        self.tag = tag

    def compile(self, args: list[Wire]) -> list[Wire]:
        [either] = args
        if self.tag == 0:
            out = build_unwrap_left(
                self.builder, either, "Either.unwrap_left: value is `right`"
            )
        else:
            out = build_unwrap_right(
                self.builder, either, "Either.unwrap_right: value is `left`"
            )
        return list(out)



"""Type checking code for top-level and nested function definitions.

For top-level functions, we take the `DefinedFunction` containing the `ast.FunctionDef`
node straight from the Python AST. We build a CFG, check it, and return a
`CheckedFunction` containing a `CheckedCFG` with type annotations.
"""

import ast
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from guppylang_internals.ast_util import return_nodes_in_ast, with_loc
from guppylang_internals.cfg.bb import BB
from guppylang_internals.cfg.builder import CFGBuilder
from guppylang_internals.checker.cfg_checker import CheckedCFG, check_cfg
from guppylang_internals.checker.core import Context, Globals, Place, Variable
from guppylang_internals.checker.errors.generic import UnsupportedError
from guppylang_internals.definition.common import DefId
from guppylang_internals.diagnostic import Error, Help, Note
from guppylang_internals.engine import DEF_STORE, ENGINE
from guppylang_internals.error import GuppyError
from guppylang_internals.experimental import check_capturing_closures_enabled
from guppylang_internals.nodes import CheckedNestedFunctionDef, NestedFunctionDef
from guppylang_internals.tys.parsing import parse_function_io_types
from guppylang_internals.tys.ty import FunctionType, InputFlags, NoneType

if sys.version_info >= (3, 12):
    from guppylang_internals.tys.parsing import parse_parameter

if TYPE_CHECKING:
    from guppylang_internals.tys.param import Parameter


@dataclass(frozen=True)
class IllegalAssignError(Error):
    title: ClassVar[str] = "Illegal assignment"
    span_label: ClassVar[str] = (
        "Variable `{var}` may not be assigned to since `{var}` is captured from an "
        "outer scope"
    )
    var: str

    @dataclass(frozen=True)
    class DefHint(Note):
        span_label: ClassVar[str] = "`{var}` defined here"
        var: str


@dataclass(frozen=True)
class MissingArgAnnotationError(Error):
    title: ClassVar[str] = "Missing type annotation"
    span_label: ClassVar[str] = "Argument requires a type annotation"


@dataclass(frozen=True)
class MissingReturnAnnotationError(Error):
    title: ClassVar[str] = "Missing type annotation"
    span_label: ClassVar[str] = "Return type must be annotated"

    @dataclass(frozen=True)
    class ReturnNone(Help):
        message: ClassVar[str] = (
            "Looks like `{func}` doesn't return anything. Consider annotating it with "
            "`-> None`."
        )
        func: str


def check_global_func_def(
    func_def: ast.FunctionDef, ty: FunctionType, globals: Globals
) -> CheckedCFG[Place]:
    """Type checks a top-level function definition."""
    args = func_def.args.args
    returns_none = isinstance(ty.output, NoneType)
    assert ty.input_names is not None

    cfg = CFGBuilder().build(func_def.body, returns_none, globals)
    inputs = [
        Variable(x, inp.ty, loc, inp.flags, is_func_input=True)
        for x, inp, loc in zip(ty.input_names, ty.inputs, args, strict=True)
        # Comptime inputs are turned into generic args, so are not included here
        if InputFlags.Comptime not in inp.flags
    ]
    generic_params = {
        param.name: param.with_idx(i) for i, param in enumerate(ty.params)
    }
    return check_cfg(cfg, inputs, ty.output, generic_params, func_def.name, globals)


def check_nested_func_def(
    func_def: NestedFunctionDef, bb: BB, ctx: Context
) -> CheckedNestedFunctionDef:
    """Type checks a local (nested) function definition."""
    func_ty = check_signature(func_def, ctx.globals)
    assert func_ty.input_names is not None

    if func_ty.parametrized:
        raise GuppyError(
            UnsupportedError(func_def, "Nested generic function definitions")
        )

    # We've already built the CFG for this function while building the CFG of the
    # enclosing function
    cfg = func_def.cfg

    # Find captured variables
    parent_cfg = bb.containing_cfg
    def_ass_before = set(func_ty.input_names) | ctx.locals.keys()
    maybe_ass_before = def_ass_before | parent_cfg.maybe_ass_before[bb]
    inout_vars = inout_var_names(func_ty)
    cfg.analyze(def_ass_before, maybe_ass_before, inout_vars)
    captured = {
        x: (ctx.locals[x], using_bb.vars.used[x])
        for x, using_bb in cfg.live_before[cfg.entry_bb].items()
        if x not in func_ty.input_names and x in ctx.locals
    }

    # Capturing closures are an experimental features since they aren't supported
    # further down the stack yet
    if captured:
        _, loc = captured[next(iter(captured.keys()))]
        check_capturing_closures_enabled(loc)

    # Captured variables may never be assigned to
    for bb in cfg.bbs:
        for v, _ in captured.values():
            x = v.name
            if x in bb.vars.assigned:
                err = IllegalAssignError(bb.vars.assigned[x], x)
                err.add_sub_diagnostic(IllegalAssignError.DefHint(v.defined_at, x))
                raise GuppyError(err)

    # Construct inputs for checking the body CFG
    inputs = [v for v, _ in captured.values()] + [
        Variable(x, inp.ty, func_def.args.args[i], inp.flags, is_func_input=True)
        for i, (x, inp) in enumerate(
            zip(func_ty.input_names, func_ty.inputs, strict=True)
        )
        # Comptime inputs are turned into generic args, so are not included here
        if InputFlags.Comptime not in inp.flags
    ]
    def_id = DefId.fresh()
    globals = ctx.globals

    # Check if the body contains a free (recursive) occurrence of the function name.
    # By checking if the name is free at the entry BB, we avoid false positives when
    # a user shadows the name with a local variable
    if func_def.name in cfg.live_before[cfg.entry_bb]:
        if not captured:
            # If there are no captured vars, we treat the function like a global name
            from guppylang.defs import GuppyDefinition
            from guppylang_internals.definition.function import ParsedFunctionDef

            func = ParsedFunctionDef(def_id, func_def.name, func_def, func_ty, None)
            DEF_STORE.register_def(func, None)
            ENGINE.parsed[def_id] = func
            globals.f_locals[func_def.name] = GuppyDefinition(func)
        else:
            # Otherwise, we treat it like a local name
            inputs.append(Variable(func_def.name, func_def.ty, func_def))

    checked_cfg = check_cfg(cfg, inputs, func_ty.output, {}, func_def.name, globals)
    checked_def = CheckedNestedFunctionDef(
        def_id,
        checked_cfg,
        func_ty,
        captured,
        name=func_def.name,
        args=func_def.args,
        body=func_def.body,
        decorator_list=func_def.decorator_list,
        returns=func_def.returns,
        type_comment=func_def.type_comment,
    )
    return with_loc(func_def, checked_def)


def check_signature(func_def: ast.FunctionDef, globals: Globals) -> FunctionType:
    """Checks the signature of a function definition and returns the corresponding
    Guppy type."""
    if len(func_def.args.posonlyargs) != 0:
        raise GuppyError(
            UnsupportedError(func_def.args.posonlyargs[0], "Positional-only parameters")
        )
    if len(func_def.args.kwonlyargs) != 0:
        raise GuppyError(
            UnsupportedError(func_def.args.kwonlyargs[0], "Keyword-only parameters")
        )
    if func_def.args.vararg is not None:
        raise GuppyError(UnsupportedError(func_def.args.vararg, "Variadic args"))
    if func_def.args.kwarg is not None:
        raise GuppyError(UnsupportedError(func_def.args.kwarg, "Keyword args"))
    if func_def.args.defaults:
        raise GuppyError(
            UnsupportedError(func_def.args.defaults[0], "Default arguments")
        )
    if func_def.returns is None:
        err = MissingReturnAnnotationError(func_def)
        # TODO: Error location is incorrect
        if all(r.value is None for r in return_nodes_in_ast(func_def)):
            err.add_sub_diagnostic(
                MissingReturnAnnotationError.ReturnNone(None, func_def.name)
            )
        raise GuppyError(err)

    # Prepopulate parameter mapping when using Python 3.12 style generic syntax
    param_var_mapping: dict[str, Parameter] = {}
    if sys.version_info >= (3, 12):
        for i, param_node in enumerate(func_def.type_params):
            param = parse_parameter(param_node, i, globals)
            param_var_mapping[param.name] = param

    input_nodes = []
    input_names = []
    for inp in func_def.args.args:
        ty_ast = inp.annotation
        if ty_ast is None:
            raise GuppyError(MissingArgAnnotationError(inp))
        input_nodes.append(ty_ast)
        input_names.append(inp.arg)
    inputs, output = parse_function_io_types(
        input_nodes,
        func_def.returns,
        input_names,
        func_def,
        globals,
        param_var_mapping,
        True,
    )
    return FunctionType(
        inputs,
        output,
        input_names,
        sorted(param_var_mapping.values(), key=lambda v: v.idx),
    )


def parse_function_with_docstring(
    func_ast: ast.FunctionDef,
) -> tuple[ast.FunctionDef, str | None]:
    """Check if the first line of a function is a docstring.

    If it is, return the function with the docstring removed, plus the docstring.
    Else, return the original function and `None`
    """
    docstring = None
    match func_ast.body:
        case [doc, *xs]:
            if (
                isinstance(doc, ast.Expr)
                and isinstance(doc.value, ast.Constant)
                and isinstance(doc.value.value, str)
            ):
                docstring = doc.value.value
                func_ast.body = xs
        case _:
            pass
    return func_ast, docstring


def inout_var_names(func_ty: FunctionType) -> list[str]:
    """Returns the names of all borrowed arguments in a function type."""
    assert func_ty.input_names is not None
    return [
        x
        for inp, x in zip(func_ty.inputs, func_ty.input_names, strict=True)
        if InputFlags.Inout in inp.flags
    ]



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit


@guppy.struct
class MyStruct:
    q1: qubit
    q2: qubit


@guppy
def foo() -> MyStruct:
    return MyStruct(qubit(), qubit())


@guppy
def bar() -> qubit:
    return foo().q1


bar.compile()



from guppylang import guppy
from guppylang_internals.decorator import custom_type
from guppylang_internals.tys.ty import NoneType

@custom_type(lambda _, ctx: NoneType().to_hugr(ctx))
class MyType:
    """A type where the `__iter__` method has the wrong signature."""

    @guppy.declare
    def __iter__(self: "MyType", x: int) -> "MyType":
        ...


@guppy
def test(x: MyType) -> None:
    for _ in x:
        pass


test.compile()



from abc import ABC

from hugr import Wire, ops
from hugr import tys as ht

from guppylang_internals.definition.custom import (
    CustomCallCompiler,
    CustomInoutCallCompiler,
)
from guppylang_internals.definition.value import CallReturnWires
from guppylang_internals.error import InternalGuppyError
from guppylang_internals.std._internal.compiler.prelude import (
    build_expect_none,
    build_unwrap,
)
from guppylang_internals.std._internal.compiler.tket_bool import (
    OPAQUE_FALSE,
    OPAQUE_TRUE,
)
from guppylang_internals.tys.arg import TypeArg


class OptionCompiler(CustomInoutCallCompiler, ABC):
    """Abstract base class for compilers for `Option` methods."""

    @property
    def option_ty(self) -> ht.Option:
        match self.type_args:
            case [TypeArg(ty)]:
                return ht.Option(ty.to_hugr(self.ctx))
            case _:
                raise InternalGuppyError("Invalid type args for Option op")


class OptionConstructor(OptionCompiler, CustomCallCompiler):
    """Compiler for the `Option` constructors `nothing` and `some`."""

    def __init__(self, tag: int):
        self.tag = tag

    def compile(self, args: list[Wire]) -> list[Wire]:
        return [self.builder.add_op(ops.Tag(self.tag, self.option_ty), *args)]


class OptionTestCompiler(OptionCompiler):
    """Compiler for the `Option.is_nothing` and `Option.is_some` methods."""

    def __init__(self, tag: int):
        self.tag = tag

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [opt] = args
        cond = self.builder.add_conditional(opt)
        for i in [0, 1]:
            with cond.add_case(i) as case:
                val = OPAQUE_TRUE if i == self.tag else OPAQUE_FALSE
                opt = case.add_op(ops.Tag(i, self.option_ty), *case.inputs())
                case.set_outputs(case.load(val), opt)
        [res, opt] = cond.outputs()
        return CallReturnWires(regular_returns=[res], inout_returns=[opt])


class OptionUnwrapCompiler(OptionCompiler, CustomCallCompiler):
    """Compiler for the `Option.unwrap` method."""

    def compile(self, args: list[Wire]) -> list[Wire]:
        [opt] = args
        err = "Option.unwrap: value is `Nothing`"
        return list(build_unwrap(self.builder, opt, err).outputs())


class OptionUnwrapNothingCompiler(OptionCompiler, CustomCallCompiler):
    """Compiler for the `Option.unwrap_nothing` method."""

    def compile(self, args: list[Wire]) -> list[Wire]:
        [opt] = args
        err = "Option.unwrap: value is `Some`"
        return list(build_expect_none(self.builder, opt, err).outputs())



from guppylang.decorator import guppy


class Generic:
    """Mock of the `typing.Generic` class to allow us to inherit from `Generic` multiple
    times.
    """

    def __class_getitem__(cls, item):
        return cls


T = guppy.type_var("T")


@guppy.struct
class MyStruct[S](Generic[T]):
    pass


MyStruct.check()



from dataclasses import dataclass
from typing import ClassVar

from guppylang_internals.checker.core import Place
from guppylang_internals.diagnostic import Error, Help, Note
from guppylang_internals.tys.ty import FunctionType


@dataclass(frozen=True)
class IllegalComptimeExpressionError(Error):
    title: ClassVar[str] = "Unsupported Python expression"
    span_label: ClassVar[str] = "Expression of type `{python_ty}` is not supported"
    python_ty: type


@dataclass(frozen=True)
class ComptimeExprNotCPythonError(Error):
    title: ClassVar[str] = "Not running CPython"
    span_label: ClassVar[str] = "Comptime expressions are only supported in CPython"


@dataclass(frozen=True)
class ComptimeExprNotStaticError(Error):
    title: ClassVar[str] = "Not compile-time evaluatable"
    span_label: ClassVar[str] = (
        "Guppy variable `{guppy_var}` cannot be accessed in a comptime expression"
    )
    guppy_var: str


@dataclass(frozen=True)
class ComptimeExprEvalError(Error):
    title: ClassVar[str] = "Python error"
    span_label: ClassVar[str] = (
        "Error occurred while evaluating this comptime expression"
    )
    message: ClassVar[str] = "Traceback printed below:\n\n{err}"
    err: str


@dataclass(frozen=True)
class ComptimeExprIncoherentListError(Error):
    title: ClassVar[str] = "Unsupported list"
    span_label: ClassVar[str] = "List contains elements with different types"


@dataclass(frozen=True)
class TketNotInstalled(Error):
    title: ClassVar[str] = "Tket not installed"
    span_label: ClassVar[str] = (
        "Experimental pytket compatibility requires `tket` to be installed"
    )

    @dataclass(frozen=True)
    class InstallInstruction(Help):
        message: ClassVar[str] = "Install tket: `pip install tket`"


@dataclass(frozen=True)
class PytketSignatureMismatch(Error):
    title: ClassVar[str] = "Signature mismatch"
    span_label: ClassVar[str] = "Signature `{name}` doesn't match provided circuit"
    name: str

    @dataclass(frozen=True)
    class TypeHint(Note):
        message: ClassVar[str] = "Expected circuit signature is `{circ_sig}`"
        circ_sig: FunctionType


@dataclass(frozen=True)
class ComptimeUnknownError(Error):
    title: ClassVar[str] = "Not known at compile-time"
    span_label: ClassVar[str] = "Value of this {thing} must be known at compile-time"
    thing: str

    @dataclass(frozen=True)
    class InputHint(Help):
        span_label: ClassVar[str] = (
            "`{place.root}` is a function argument, so its value is not statically "
            "known. Consider turning `{place.root}` into a comptime argument: "
            "`{place.root}: @comptime`"
        )
        place: Place

    @dataclass(frozen=True)
    class VariableHint(Note):
        span_label: ClassVar[str] = (
            "`{place.root}` is a dynamic variable, so its value is not statically known"
        )
        place: Place

    @dataclass(frozen=True)
    class FallbackHint(Note):
        span_label: ClassVar[str] = (
            "This expression involves a dynamic computation, so its value is not "
            "statically known"
        )

    @dataclass(frozen=True)
    class Feedback(Note):
        message: ClassVar[str] = (
            "We are currently investigating ways to make Guppy's compile-time "
            "reasoning smarter. Please leave your feedback at "
            "https://github.com/CQCL/guppylang/discussions/987"
        )



from dataclasses import dataclass

from hugr import val
from tket_exts import (
    debug,
    futures,
    opaque_bool,
    qsystem,
    qsystem_random,
    qsystem_utils,
    quantum,
    result,
    rotation,
    wasm,
)

BOOL_EXTENSION = opaque_bool()
DEBUG_EXTENSION = debug()
FUTURES_EXTENSION = futures()
QSYSTEM_EXTENSION = qsystem()
QSYSTEM_RANDOM_EXTENSION = qsystem_random()
QSYSTEM_UTILS_EXTENSION = qsystem_utils()
QUANTUM_EXTENSION = quantum()
RESULT_EXTENSION = result()
ROTATION_EXTENSION = rotation()
WASM_EXTENSION = wasm()

TKET_EXTENSIONS = [
    BOOL_EXTENSION,
    DEBUG_EXTENSION,
    FUTURES_EXTENSION,
    QSYSTEM_EXTENSION,
    QSYSTEM_RANDOM_EXTENSION,
    QSYSTEM_UTILS_EXTENSION,
    QUANTUM_EXTENSION,
    RESULT_EXTENSION,
    ROTATION_EXTENSION,
    WASM_EXTENSION,
]


@dataclass(frozen=True)
class ConstWasmModule(val.ExtensionValue):
    """Python wrapper for the tket ConstWasmModule type"""

    wasm_file: str
    wasm_hash: int

    def to_value(self) -> val.Extension:
        ty = WASM_EXTENSION.get_type("module").instantiate([])

        name = "tket.wasm.ConstWasmModule"
        payload = {"name": self.wasm_file, "hash": self.wasm_hash}
        return val.Extension(name, typ=ty, val=payload, extensions=["tket.wasm"])

    def __str__(self) -> str:
        return (
            f"ConstWasmModule(wasm_file={self.wasm_file}, wasm_hash={self.wasm_hash})"
        )



from guppylang import guppy
from guppylang.std.collections import (
    Stack,
    empty_stack,
    PriorityQueue,
    empty_priority_queue,
)


def test_stack(run_int_fn) -> None:
    @guppy
    def main() -> int:
        stack: Stack[int, 10] = empty_stack()
        for i in range(10):
            stack = stack.push(i)
        s = 0
        i = 1
        while len(stack) > 0:
            x, stack = stack.pop()
            s += x * i
            i += 1
        return s

    run_int_fn(main, sum((i + 1) * x for i, x in enumerate(reversed(list(range(10))))))


def test_priority_queue(run_int_fn) -> None:
    @guppy
    def main() -> int:
        pq: PriorityQueue[int, 10] = empty_priority_queue()
        for i in range(10):
            # values are in order, priority is reversed
            pq = pq.push(i, 9 - i)
        s = 0
        multiplier = 1
        while len(pq) > 0:
            priority, value, pq = pq.pop()
            # use multiplier to ensure the correct order
            s += value * multiplier
            multiplier += 1
        return s

    run_int_fn(
        main,
        # multiplier * value for ordered values in priority queue
        sum((m + 1) * v for m, v in enumerate(reversed(list(range(10))))),
    )



from guppylang.decorator import guppy
from guppylang.std.builtins import array, owned


@guppy.declare
def use(q: array[int, 3] @owned) -> None: ...


@guppy
def test(q: array[int, 3] @owned) -> None:
    p = q
    use(q)


test.compile()


from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.struct
class MyStruct:
    q1: qubit
    q2: qubit


@guppy
def foo(s: MyStruct @owned) -> qubit:
    return s.q1


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit, measure


@guppy.struct
class MyStruct1:
    x: "MyStruct2"


@guppy.struct
class MyStruct2:
    q1: qubit
    q2: qubit


@guppy
def foo(s: MyStruct1 @owned) -> MyStruct1:
    measure(s.x.q1)
    return MyStruct1(s.x)


foo.compile()



from guppylang.std.debug import state_result
from guppylang.std.quantum import discard, qubit
from tests.util import compile_guppy

@compile_guppy
def main(y: bool) -> None:
    q1 = qubit()
    state_result("foo" if y else "bar", q1)
    discard(q1)



from guppylang import guppy
from guppylang.std.builtins import frozenarray, panic, py


def test_len(run_int_fn):
    @guppy
    def foo(xs: frozenarray[int, 42]) -> int:
        return len(xs)

    @guppy
    def main() -> int:
        xs = py(list(range(42)))
        return foo(xs)

    run_int_fn(main, 42)


def test_subscript(run_int_fn):
    @guppy
    def foo(xs: frozenarray[int, 42]) -> int:
        return xs[10] + xs[10]

    @guppy
    def main() -> int:
        xs = py(list(range(42)))
        return foo(xs)

    run_int_fn(main, 20)


def test_iter(run_int_fn):
    @guppy
    def foo(xs: frozenarray[int, 42]) -> int:
        s = 0
        for x in xs:
            s += x
        return s

    @guppy
    def main() -> int:
        xs = py(list(range(42)))
        return foo(xs)

    run_int_fn(main, sum(range(42)))


def test_alias(run_int_fn):
    @guppy
    def foo(xs: frozenarray[int, 42]) -> int:
        ys = xs
        return xs[0] + ys[1]

    @guppy
    def main() -> int:
        xs = py(list(range(42)))
        return foo(xs)

    run_int_fn(main, 1)


def test_mutable_copy(run_int_fn):
    @guppy
    def foo(xs: frozenarray[int, 42]) -> int:
        ys = xs.mutable_copy()
        s = 0
        for i in range(42):
            if xs[i] != ys[i]:
                panic("Mismatch")
            s += xs[i]
            s += ys[i]
        return s

    @guppy
    def main() -> int:
        xs = py(list(range(42)))
        return foo(xs)

    run_int_fn(main, 2 * sum(range(42)))


def test_nested_subscript(run_int_fn):
    @guppy
    def foo(xs: frozenarray[frozenarray[int, 2], 2]) -> int:
        return xs[0][0] + xs[0][1] + xs[1][0] + xs[1][1]

    @guppy
    def main() -> int:
        xs = py([[1, 2], [3, 4]])
        return foo(xs)

    run_int_fn(main, 1 + 2 + 3 + 4)


def test_nested_iter(run_int_fn):
    @guppy
    def foo(xss: frozenarray[frozenarray[int, 10], 42]) -> int:
        s = 0
        for xs in xss:
            for x in xs:
                s += x
        return s

    xss = [[i * j for j in range(10)] for i in range(42)]

    @guppy
    def main() -> int:
        return foo(py(xss))

    run_int_fn(main, sum([sum(xs) for xs in xss]))


def test_nested_struct(validate):
    @guppy.struct
    class S:
        xs: frozenarray[int, 10]
        y: int

    @guppy
    def foo(xss: frozenarray[S, 42]) -> int:
        s = xss[0].y + xss[0].xs[0]
        for ss in xss:
            for x in ss.xs:
                s += x
        return s

    validate(foo.compile())



from hugr import Hugr
from hugr.package import Package, PackagePointer

from pathlib import Path
import pytest
from typing import Any

from selene_hugr_qis_compiler import check_hugr

from guppylang.defs import GuppyDefinition


@pytest.fixture(scope="session")
def export_test_cases_dir(request):
    r = request.config.getoption("--export-test-cases")
    if r and not r.exists():
        r.mkdir(parents=True)
    return r


@pytest.fixture(scope="session")
def validate(request, export_test_cases_dir: Path):
    def validate_impl(package: Package | PackagePointer | Hugr, name=None):
        if isinstance(package, PackagePointer):
            package = package.package
        if isinstance(package, Hugr):
            package = Package([package])
        # Validate via the json encoding
        package_bytes = package.to_bytes()

        if export_test_cases_dir:
            file_name = f"{request.node.name}{f'_{name}' if name else ''}.hugr"
            export_file = export_test_cases_dir / file_name
            export_file.write_bytes(package_bytes)

        check_hugr(package_bytes)

    return validate_impl


class LLVMException(Exception):
    pass


def _emulate_fn(is_flt: bool = False):
    """Use selene to emulate a Guppy function."""
    from guppylang.decorator import guppy
    from guppylang.std.builtins import result

    def f(f: GuppyDefinition, expected: Any, args: list[Any] | None = None):
        args = args or []

        @guppy.comptime
        def int_entry() -> None:
            o: int = f(*args)
            result("_test_output", o)

        @guppy.comptime
        def flt_entry() -> None:
            o: float = f(*args)
            result("_test_output", o)

        entry = flt_entry if is_flt else int_entry
        res = entry.emulator(0).coinflip_sim().with_seed(42).run()
        num = next(v for k, v in res.results[0].entries if k == "_test_output")
        if num != expected:
            raise LLVMException(
                f"Expected value ({expected}) doesn't match actual value ({num})"
            )

    return f


@pytest.fixture
def run_int_fn():
    """Emulate an integer function using the Guppy emulator."""
    return _emulate_fn(is_flt=False)


@pytest.fixture
def run_float_fn_approx():
    """Like run_int_fn, but takes optional additional parameters `rel`, `abs`
    and `nan_ok` as per `pytest.approx`."""
    run_fn = _emulate_fn(is_flt=True)

    def run_approx(
        f: GuppyDefinition,
        expected: float,
        args: list[Any] | None = None,
        *,
        rel: float | None = None,
        abs: float | None = None,
        nan_ok: bool = False,
    ):
        return run_fn(
            f,
            pytest.approx(expected, rel=rel, abs=abs, nan_ok=nan_ok),
            args,
        )

    return run_approx



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.declare
def new_qubit() -> qubit:
    ...


@guppy
def foo(q: qubit @owned) -> qubit:
    q = new_qubit()
    return q


foo.compile()



import ast
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import ClassVar

from guppylang_internals.checker.core import Globals
from guppylang_internals.definition.common import CompiledDef, Definition, ParsableDef
from guppylang_internals.diagnostic import Error
from guppylang_internals.error import GuppyError, InternalGuppyError
from guppylang_internals.span import SourceMap
from guppylang_internals.tys.param import ConstParam, Parameter, TypeParam
from guppylang_internals.tys.ty import Type


@dataclass(frozen=True)
class LinearConstVarError(Error):
    title: ClassVar[str] = "Invalid const variable"
    span_label: ClassVar[str] = (
        "Const variable `{name}` is not allowed have {thing} type `{ty}`"
    )
    name: str
    ty: Type

    @property
    def thing(self) -> str:
        return "non-copyable" if not self.ty.copyable else "non-droppable"


class ParamDef(Definition):
    """Abstract base class for type parameter definitions."""

    @abstractmethod
    def to_param(self, idx: int) -> Parameter:
        """Creates a parameter from this definition."""


@dataclass(frozen=True)
class TypeVarDef(ParamDef, CompiledDef):
    """A type variable definition."""

    must_be_copyable: bool
    must_be_droppable: bool

    description: str = field(default="type variable", init=False)

    def to_param(self, idx: int) -> TypeParam:
        """Creates a parameter from this definition."""
        return TypeParam(idx, self.name, self.must_be_copyable, self.must_be_droppable)


@dataclass(frozen=True)
class RawConstVarDef(ParamDef, ParsableDef):
    """A constant variable definition whose type is not parsed yet."""

    type_ast: ast.expr
    description: str = field(default="const variable", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "ConstVarDef":
        from guppylang_internals.tys.parsing import type_from_ast

        ty = type_from_ast(self.type_ast, globals, {})
        if not ty.copyable or not ty.droppable:
            raise GuppyError(LinearConstVarError(self.type_ast, self.name, ty))
        return ConstVarDef(self.id, self.name, self.defined_at, ty)

    def to_param(self, idx: int) -> Parameter:
        raise InternalGuppyError(
            "RawConstVarDef.to_param: Parameter conversion only possible after parsing"
        )


@dataclass(frozen=True)
class ConstVarDef(ParamDef, CompiledDef):
    """A constant variable definition."""

    ty: Type

    description: str = field(default="const variable", init=False)

    def to_param(self, idx: int) -> ConstParam:
        """Creates a parameter from this definition."""
        return ConstParam(idx, self.name, self.ty)



from guppylang import guppy
from guppylang_internals.decorator import custom_type
from guppylang_internals.tys.ty import NoneType

@custom_type(lambda _, ctx: NoneType().to_hugr(ctx))
class MyIter:
    """An iterator where the `__next__` method has the wrong signature."""

    @guppy.declare
    def __next__(self: "MyIter") -> tuple["MyIter", float]:
        ...

    @guppy.declare
    def __hasnext__(self: "MyIter") -> tuple[bool, "MyIter"]:
        ...

    @guppy.declare
    def __end__(self: "MyIter") -> None:
        ...


@custom_type(lambda _, ctx: NoneType().to_hugr(ctx))
class MyType:
    """Type that produces the iterator above."""

    @guppy.declare
    def __iter__(self: "MyType") -> MyIter:
        ...


@guppy
def test(x: MyType) -> None:
    for _ in x:
        pass


test.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import array, owned
from guppylang.std.quantum import qubit


@guppy
def main(xs: array[qubit, 2] @ owned) -> array[int, 2]:
   ys = xs.copy()
   return ys

main.compile()



import pathlib
import pytest

from guppylang_internals.tracing.state import reset_state
from tests.error.util import run_error_test

path = pathlib.Path(__file__).parent.resolve() / "tracing_errors"
files = [
    x
    for x in path.iterdir()
    if x.is_file() and x.suffix == ".py" and x.name != "__init__.py"
]

# Turn paths into strings, otherwise pytest doesn't display the names
files = [str(f) for f in files]


@pytest.mark.parametrize("file", files)
def test_tracing_errors(file, capsys, snapshot):
    # Reset the tracing state by hand since the previous test catches the exception so
    # it's not reset
    reset_state()
    run_error_test(file, capsys, snapshot)



from guppylang.decorator import guppy
from guppylang.std.futures import Future
from guppylang.std.lang import owned


def test_read(validate):
    @guppy
    def main(x: Future[int], y: Future[bool] @ owned) -> int:
        z = x.copy()
        if y.read():
            return z.read()
        z.discard()
        return 0

    validate(main.compile())



from guppylang.decorator import guppy
from guppylang.std.builtins import array

n = guppy.nat_var("n")


@guppy.declare
def foo(x: array[int, n], y: array[int, n]) -> None:
    ...


@guppy
def main(x: array[int, 42], y: array[int, 43]) -> None:
    foo(x, y)


main.compile()



import pytest

from guppylang.decorator import guppy
from guppylang_internals.error import GuppyError


def test_func_redefinition(validate):
    @guppy
    def test() -> bool:
        return 5  # Type error on purpose

    @guppy
    def test() -> bool:  # noqa: F811
        return False

    validate(test.compile())


def test_method_redefinition(validate):
    @guppy.struct
    class Test:
        x: int

        @guppy
        def foo(self: "Test") -> int:
            return 1.0  # Type error on purpose

        @guppy
        def foo(self: "Test") -> int:  # noqa: F811
            return 1  # Type error on purpose

    @guppy
    def main(t: Test) -> int:
        return t.foo()

    validate(main.compile())


def test_redefine_after_error(validate):
    @guppy.struct
    class Foo:
        x: int

    @guppy
    def foo() -> int:
        return y  # noqa: F821

    with pytest.raises(GuppyError):
        foo.compile()

    @guppy.struct
    class Foo:  # noqa: F811
        x: int

    @guppy
    def foo(f: Foo) -> int:
        return f.x

    validate(foo.compile())


@pytest.mark.skip("See https://github.com/CQCL/guppylang/issues/456")
def test_struct_redefinition(validate):
    @guppy.struct
    class Test:
        x: "blah"  # Non-existing type  # noqa: F821

    @guppy.struct
    class Test:  # noqa: F811
        y: int

    @guppy
    def main(x: int) -> Test:
        return Test(x)

    validate(main.compile())


@pytest.mark.skip("See https://github.com/CQCL/guppylang/issues/456")
def test_struct_method_redefinition(validate):
    @guppy.struct
    class Test:
        x: int

        @guppy
        def foo(self: "Test") -> int:
            return 1.0  # Type error on purpose

    @guppy.struct
    class Test:  # noqa: F811
        y: int

        @guppy
        def bar(self: "Test") -> int:
            return self.y

    @guppy
    def main(x: int) -> int:
        return Test(x).bar()

    validate(main.compile())



from guppylang.decorator import guppy


@guppy.struct
class MyStruct:
    @guppy
    def x(self: "MyStruct") -> int:
        return 0

    x: int


@guppy
def main(s: MyStruct) -> None:
    s.x


main.compile()
